/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Machine Perception and Intelligent    |
   |      Robotics Lab, University of Malaga (Spain).                          |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/precomp_core.h>  // Only for precomp. headers, include all libmrpt-core headers.


#include <mrpt/slam/CRangeBearingKFSLAM2D.h>
#include <mrpt/poses/CPosePDF.h>
#include <mrpt/poses/CPosePDFGaussian.h>
#include <mrpt/slam/CActionRobotMovement3D.h>

#include <mrpt/math/utils.h>
#include <mrpt/math/CMatrix.h>
#include <mrpt/math/CMatrixD.h>
#include <mrpt/utils/CTicTac.h>

using namespace mrpt::slam;
using namespace mrpt::poses;
using namespace mrpt::utils;
using namespace std;

/*---------------------------------------------------------------
							Constructor
  ---------------------------------------------------------------*/
CRangeBearingKFSLAM2D::CRangeBearingKFSLAM2D( ) :
	options(),
	m_action(),m_SF(),
	m_IDs(),
	m_IDs_inverse(),
	m_SFs()
{
	reset();
}

/*---------------------------------------------------------------
							reset
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM2D::reset()
{
	m_action.clear_unique();
	m_SF.clear_unique();
	m_IDs.clear();
	m_IDs_inverse.clear();
	m_SFs.clear();

	// INIT KF STATE
	m_xkk.assign(3,0);	// State: 3D pose (x,y,phi)

	// Initial cov:
	m_pkk.setSize(3,3);
	m_pkk.zeros();
}

/*---------------------------------------------------------------
							Destructor
  ---------------------------------------------------------------*/
CRangeBearingKFSLAM2D::~CRangeBearingKFSLAM2D()
{
}

/*---------------------------------------------------------------
							getCurrentRobotPose
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM2D::getCurrentRobotPose(
	CPosePDFGaussian      &out_robotPose ) const
{
	MRPT_TRY_START

	ASSERT_(m_xkk.size()>=3);

	// Set 6D pose mean:
	out_robotPose.mean = CPose2D(m_xkk[0],m_xkk[1],m_xkk[2]);

	// and cov:
	CMatrixTemplateNumeric<KFTYPE>  COV(3,3);
	m_pkk.extractMatrix(0,0,COV);
	out_robotPose.cov = COV;

	MRPT_TRY_END
}


/*---------------------------------------------------------------
							getCurrentState
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM2D::getCurrentState(
	CPosePDFGaussian      &out_robotPose,
	std::vector<TPoint2D>  &out_landmarksPositions,
	std::map<unsigned int,CLandmark::TLandmarkID> &out_landmarkIDs,
	CVectorDouble      &out_fullState,
	CMatrixDouble      &out_fullCovariance ) const
{
	MRPT_TRY_START

	ASSERT_(m_xkk.size()>=3);

	// Set 6D pose mean:
	out_robotPose.mean = CPose2D(m_xkk[0],m_xkk[1],m_xkk[2]);

	// and cov:
	CMatrixTemplateNumeric<KFTYPE>  COV(3,3);
	m_pkk.extractMatrix(0,0,COV);
	out_robotPose.cov = COV;


	// Landmarks:
	ASSERT_( ((m_xkk.size() - 3) % 2)==0 );
	size_t i,nLMs = (m_xkk.size() - 3) / 2;
	out_landmarksPositions.resize(nLMs);
	for (i=0;i<nLMs;i++)
	{
		out_landmarksPositions[i].x = m_xkk[3+i*2+0];
		out_landmarksPositions[i].y = m_xkk[3+i*2+1];
	} // end for i

	// IDs:
	out_landmarkIDs = m_IDs_inverse;

    // Full state:
    out_fullState.resize( m_xkk.size() );
    for (i=0;i<m_xkk.size();i++)
		out_fullState[i] = m_xkk[i];

	// Full cov:
	out_fullCovariance = m_pkk;

	MRPT_TRY_END
}


/*---------------------------------------------------------------
						processActionObservation
  ---------------------------------------------------------------*/
void  CRangeBearingKFSLAM2D::processActionObservation(
    CActionCollectionPtr	&action,
    CSensoryFramePtr		&SF )
{
	MRPT_TRY_START

	m_action = action;
	m_SF = SF;

	// Sanity check:
	ASSERT_( m_IDs.size() == (m_pkk.getColCount() - get_vehicle_size() )/get_feature_size() );
	ASSERT_( m_IDs.size() == m_IDs_inverse.size() );

	// ===================================================================================================================
	// Here's the meat!: Call the main method for the KF algorithm, which will call all the callback methods as required:
	// ===================================================================================================================
	runOneKalmanIteration();


	// =============================================================
	//  ADD TO SFs SEQUENCE
	// =============================================================
	if (options.create_simplemap)
	{
		CPosePDFGaussianPtr	auxPosePDF = CPosePDFGaussian::Create();
		getCurrentRobotPose( *auxPosePDF );
		m_SFs.insert( auxPosePDF , SF );
	}

    MRPT_TRY_END
}


/** Must return the action vector u.
  * \param out_u The action vector which will be passed to OnTransitionModel
  */
void CRangeBearingKFSLAM2D::OnGetAction( CVectorTemplate<KFTYPE> &u )
{
	u.resize(3);

	// Get odometry estimation:
	CActionRobotMovement2DPtr actMov2D = m_action->getBestMovementEstimation();
	CActionRobotMovement3DPtr actMov3D = m_action->getActionByClass<CActionRobotMovement3D>();
	if (actMov3D)
	{
		u[0]=actMov3D->poseChange.mean.x();
		u[1]=actMov3D->poseChange.mean.y();
		u[2]=actMov3D->poseChange.mean.yaw();
	}
	else
	if (actMov2D)
	{
		CPose2D estMovMean;
		actMov2D->poseChange->getMean(estMovMean);
		u[0]=estMovMean.x();
		u[1]=estMovMean.y();
		u[2]=estMovMean.phi();
	}
	else
	{
		// Left u as zeros
		for (size_t i=0;i<3;i++) u[i]=0;
	}

}


/** This virtual function musts implement the prediction model of the Kalman filter.
 */
void  CRangeBearingKFSLAM2D::OnTransitionModel(
	const CVectorTemplate<KFTYPE> &u,
	CVectorTemplate<KFTYPE>        &xv,
	bool &out_skipPrediction )
{
	MRPT_TRY_START

	// Do not update the vehicle pose & its covariance until we have some landmakrs in the map,
	// otherwise, we are imposing a lower bound to the best uncertainty from now on:
	if (m_xkk.size() == get_vehicle_size() )
	{
		out_skipPrediction = true;
		return;
	}

	ASSERT_(xv.size()>=3);
	ASSERT_(u.size()==3); // 2D odometry

	CPose2D  robotPose(xv[0],xv[1],xv[2]);
	CPose2D  odoIncrement(u[0],u[1],u[2]);

	// Pose composition:
	robotPose = robotPose + odoIncrement;

	xv[0]=robotPose.x();
	xv[1]=robotPose.y();
	xv[2]=robotPose.phi();

	MRPT_TRY_END
}

/** This virtual function musts calculate the Jacobian F of the prediction model.
 */
void  CRangeBearingKFSLAM2D::OnTransitionJacobian(
	CMatrixTemplateNumeric<KFTYPE> & F )
{
	MRPT_TRY_START

	// The jacobian of the transition function:  dfv_dxv
	CActionRobotMovement2DPtr act2D = m_action->getBestMovementEstimation();
	CActionRobotMovement3DPtr act3D = m_action->getActionByClass<CActionRobotMovement3D>();

	if (act3D && act2D)  THROW_EXCEPTION("Both 2D & 3D odometry are present!?!?")

	TPoint2D  Ap;

	if (act3D)
	{
		Ap = TPoint2D(CPose2D(act3D->poseChange.mean));
	}
	else if (act2D)
	{
		Ap = TPoint2D(act2D->poseChange->getMeanVal());
	}
	else
	{
		// No odometry at all:
		F.unit(3); // Unit diagonal
		return;
	}

	const KFTYPE cy = cos(m_xkk[2]);
	const KFTYPE sy = sin(m_xkk[2]);

	const KFTYPE Ax = Ap.x;
	const KFTYPE Ay = Ap.y;

	F.unit(3); // Unit diagonal

	F.get_unsafe(0,2) = -Ax*sy-Ay*cy;
	F.get_unsafe(1,2) =  Ax*cy-Ay*sy;

	MRPT_TRY_END
}

/** This virtual function musts calculate de noise matrix of the prediction model.
 */
void  CRangeBearingKFSLAM2D::OnTransitionNoise( CMatrixTemplateNumeric<KFTYPE> & Q )
{
	MRPT_TRY_START

	// The uncertainty of the 2D odometry, projected from the current position:
	CActionRobotMovement2DPtr act2D = m_action->getBestMovementEstimation();
	CActionRobotMovement3DPtr act3D = m_action->getActionByClass<CActionRobotMovement3D>();

	if (act3D && act2D)  THROW_EXCEPTION("Both 2D & 3D odometry are present!?!?")

	CPosePDFGaussian odoIncr;

	if (!act3D && !act2D)
	{
		// Use constant Q:
		Q.zeros(3,3);
		ASSERT_(options.stds_Q_no_odo.size()==Q.getColCount())

		for (size_t i=0;i<3;i++)
			Q(i,i) = square( options.stds_Q_no_odo[i]);
		return;
	}
	else
	{
		if (act2D)
		{
			// 2D odometry:
			odoIncr = *act2D->poseChange;
		}
		else
		{
			// 3D odometry:
			odoIncr = CPosePDFGaussian(act3D->poseChange);
		}
	}

	odoIncr.rotateCov(m_xkk[2]);

	Q = odoIncr.cov;

	MRPT_TRY_END
}


/** Implements the observation prediction \f$ h_i(x) \f$ and the Jacobians \f$ \frac{\partial h_i}{\partial x} \f$ and (when applicable) \f$ \frac{\partial h_i}{\partial y_i} \f$.
  * \param in_z This is the same matrix returned by OnGetObservations(), passed here for reference.
  * \param in_data_association The vector returned by OnGetObservations(), passed here for reference.
  * \param in_full If set to true, all the Jacobians and predictions must be computed at once. Otherwise, just those for the observation in_obsIdx.
  * \param in_obsIdx If in_full=false, the row of the observation (in in_z and in_data_association) whose innovation & Jacobians are to be returned now.
  * \param out_innov The difference between the expected observation and the real one: \f$ \tilde{y} = z - h(x) \f$. It must be a vector of length O*N (in_full=true) or O (in_full=false), with O=get_observation_size() and N=number of rows in in_z.
  * \param out_Hx One or a vertical stack of \f$ \frac{\partial h_i}{\partial x} \f$.
  * \param out_Hy An empty matrix, or one or a vertical stack of \f$ \frac{\partial h_i}{\partial y_i} \f$.
  *
  *  out_Hx must consist of 1 (in_full=false) or N (in_full=true) vertically stacked \f$O \times V\f$ matrices, and out_Hy must consist of 1 or N \f$O \times F\f$ matrices, with:
  *  - N: number of rows in in_z (number of observed features, or 1 in general).
  *  - O: get_observation_size()
  *  - V: get_vehicle_size()
  *  - F: get_feature_size()
  */
void CRangeBearingKFSLAM2D::OnObservationModelAndJacobians(
	const CMatrixTemplateNumeric<KFTYPE> &in_z,
	const vector_int                     &data_association,
	const bool                           &in_full,
	const int                            &in_obsIdx,
	CVectorTemplate<KFTYPE>       		 &ytilde,
	CMatrixTemplateNumeric<KFTYPE>       &Hx,
	CMatrixTemplateNumeric<KFTYPE>       &Hy )
{
	MRPT_TRY_START

	// Get the sensor pose relative to the robot:
	CObservationBearingRangePtr obs = m_SF->getObservationByClass<CObservationBearingRange>();
	ASSERT_(obs);
	const CPose2D  sensorPoseOnRobot = CPose2D(obs->sensorLocationOnRobot);

	/* -------------------------------------------
       Equations, obtained using matlab, of the relative 2D position of a landmark (xi,yi), relative
          to a robot 2D pose (x0,y0,phi)
        Refer to technical report "6D EKF derivation...", 2008

        x0 y0 phi0         % Robot's 2D pose
        x0s y0s phis      % Sensor's 2D pose relative to robot
        xi yi             % Absolute 2D landmark coordinates:

        Hx : dh_dxv   -> Jacobian of the observation model wrt the robot pose
        Hy : dh_dyi   -> Jacobian of the observation model wrt each landmark mean position

        Sizes:
	     h:  Lx2
         Hx: 2Lx3
         Hy: 2Lx2

          L=# of landmarks in the map (ALL OF THEM)
	  ------------------------------------------- */
	size_t nLMs = (m_xkk.size() - get_vehicle_size() ) / get_feature_size();


	// Mean of the prior of the robot pose:
	const CPose2D  robotPose( m_xkk[0],m_xkk[1],m_xkk[2] );

	// Requested to build just one LM?
	if (!in_full)
        nLMs = 1;

    const size_t  vehicle_size = get_vehicle_size();
    const size_t  obs_size  = get_observation_size();
    const size_t  feature_size = get_feature_size();

	ytilde.resize( obs_size*nLMs,0);
    Hx.setSize    ( obs_size*nLMs, vehicle_size );
    Hy.setSize    ( obs_size*nLMs, feature_size );

    // Robot 6D pose:
    const KFTYPE x0 = m_xkk[0];
    const KFTYPE y0 = m_xkk[1];
    const KFTYPE phi0 = m_xkk[2];

    const KFTYPE cphi0 = cos(phi0);
    const KFTYPE sphi0 = sin(phi0);

    // Sensor 2D pose on robot:
    const KFTYPE x0s = sensorPoseOnRobot.x();
    const KFTYPE y0s = sensorPoseOnRobot.y();
	const KFTYPE phis = sensorPoseOnRobot.phi();

    const KFTYPE cphi0s = cos(phi0+phis);
    const KFTYPE sphi0s = sin(phi0+phis);

	const CPose2D sensorPoseAbs = robotPose + sensorPoseOnRobot;

	for (size_t i=0;i<nLMs;i++)  // nLMs = 1 if full=false!
	{
		size_t idxOfFeatureInMap;

		if (in_full)
			idxOfFeatureInMap = i;
		else
		{
			ASSERT_(data_association[in_obsIdx]>=0)  // It is an already mapped feature!
			idxOfFeatureInMap = data_association[in_obsIdx];  // Just take one LM
		}

		size_t  row_in = feature_size*idxOfFeatureInMap;
		size_t  row = obs_size*i;

		//tictac.Tic();

        // Landmark absolute 3D position in the map:
	    const KFTYPE xi = m_xkk[ vehicle_size + row_in + 0 ];
		const KFTYPE yi = m_xkk[ vehicle_size + row_in + 1 ];

		// ---------------------------------------------------
		// Generate Range, yaw and the ytilde:
		// ---------------------------------------------------
		const double Axi = xi-sensorPoseAbs.x();
		const double Ayi = yi-sensorPoseAbs.y();
		const double h_range = sqrt( square(Axi)+square(Ayi) );
		const double h_yaw   = mrpt::math::wrapToPi( atan2(Ayi,Axi) - sensorPoseAbs.phi() );

		// Fill out "ytilde": If this LM has been observed, the usual value. Otherwise, zeroes:
		int idxThisLMInZ = -1;
		for (size_t p=0;p< data_association.size();p++)
		{
			if ( data_association[p] == static_cast<int>(idxOfFeatureInMap))
			{
				idxThisLMInZ = p;
				break;
			}
		}

		if (idxThisLMInZ<0)
		{
			// Not observed:
			ytilde[row+0] =
			ytilde[row+1] = 0;
		}
		else
		{
			// Yes: compute ytilde:
			ytilde[row+0] = in_z(idxThisLMInZ,0) - h_range;
			ytilde[row+1] = math::wrapToPi( in_z(idxThisLMInZ,1) - h_yaw );
		}

		// ---------------------------------------------------
		// Generate dhi_dxv: A 2x3 block
		// ---------------------------------------------------
		const KFTYPE EXP1 = -2*yi*y0s*cphi0-2*yi*y0+2*xi*y0s*sphi0-2*xi*x0-2*xi*x0s*cphi0-2*yi*x0s*sphi0+2*y0s*y0*cphi0-2*y0s*x0*sphi0+2*y0*x0s*sphi0+square(x0)+2*x0s*x0*cphi0+square(x0s)+square(y0s)+square(xi)+square(yi)+square(y0);
		const KFTYPE sqrtEXP1_1 = KFTYPE(1)/sqrt(EXP1);

		const KFTYPE EXP2 = cphi0s*xi+sphi0s*yi-sin(phis)*y0s-y0*sphi0s-x0s*cos(phis)-x0*cphi0s;
		const KFTYPE EXP2sq = square(EXP2);

		const KFTYPE EXP3 = -sphi0s*xi+cphi0s*yi-cos(phis)*y0s-y0*cphi0s+x0s*sin(phis)+x0*sphi0s;
		const KFTYPE EXP3sq = square(EXP3);

		const KFTYPE EXP4 = KFTYPE(1)/(1+EXP3sq/EXP2sq);


        Hx(row+0,0)= (-xi-sphi0*y0s+cphi0*x0s+x0)*sqrtEXP1_1;
        Hx(row+0,1)= (-yi+cphi0*y0s+y0+sphi0*x0s)*sqrtEXP1_1;
		Hx(row+0,2)= (y0s*xi*cphi0+y0s*yi*sphi0-y0*y0s*sphi0-x0*y0s*cphi0+x0s*xi*sphi0-x0s*yi*cphi0+y0*x0s*cphi0-x0s*x0*sphi0)*sqrtEXP1_1;

        Hx(row+1,0)= (sphi0s/(EXP2)+(EXP3)/EXP2sq*cphi0s) * EXP4;
        Hx(row+1,1)= (-cphi0s/(EXP2)+(EXP3)/EXP2sq*sphi0s) * EXP4;
		Hx(row+1,2)= ((-cphi0s*xi-sphi0s*yi+y0*sphi0s+x0*cphi0s)/(EXP2)-(EXP3)/EXP2sq*(-sphi0s*xi+cphi0s*yi-y0*cphi0s+x0*sphi0s)) * EXP4;

        // ---------------------------------------------------
        // Generate dhi_dyi: A 2x2 block
        // ---------------------------------------------------
        Hy(row+0,0)= (xi+sphi0*y0s-cphi0*x0s-x0)*sqrtEXP1_1;
        Hy(row+0,1)= (yi-cphi0*y0s-y0-sphi0*x0s)*sqrtEXP1_1;

        Hy(row+1,0)= (-sphi0s/(EXP2)-(EXP3)/EXP2sq*cphi0s)* EXP4;
        Hy(row+1,1)= (cphi0s/(EXP2)-(EXP3)/EXP2sq*sphi0s) * EXP4;

	}

	MRPT_TRY_END
}


/** This is called between the KF prediction step and the update step to allow the application to employ the prior distribution of the system state in the detection of observations (data association), where applicable.
  * \param out_z A \f$N \times O\f$ matrix, with O being get_observation_size() and N the number of "observations": how many observed landmarks for a map, or just one if not applicable.
  * \param out_R A matrix of size N·OxO (O being get_observation_size() and N the number of observations). It is the covariance matrix of the sensor noise for each of the returned observations. The order must correspond to that in out_z.
  * \param out_data_association An empty vector or, where applicable, a vector where the i'th element corresponds to the position of the observation in the i'th row of out_z within the system state vector, or -1 if it is a new map element and we want to insert it at the end of this KF iteration.
  *  This method will be called just once for each complete KF iteration.
  */
void CRangeBearingKFSLAM2D::OnGetObservations(
	CMatrixTemplateNumeric<KFTYPE> &Z,
	CVectorTemplate<KFTYPE>        &R2,
	vector_int                     &data_association
	)
{
	MRPT_TRY_START

	// Observations: Z
	// ---------------------
	deque<CObservationBearingRange::TMeasurement>::const_iterator itObs;
	vector_int::iterator                                          itDA;

	CObservationBearingRangePtr obs = m_SF->getObservationByClass<CObservationBearingRange>();
	ASSERT_(obs);

	// Z: NxO  (N:#obs, O:2)
	// data_association: N

	size_t N = obs->sensedData.size();
	Z.setSize(N,2);
	data_association.clear();
	data_association.resize(N,-1);

	// For each observed LM:
	size_t row;
    for (row=0,itObs = obs->sensedData.begin(),itDA=data_association.begin();itObs!=obs->sensedData.end();itObs++,itDA++,row++)
    {
    	// Fill one row in Z:
    	Z(row,0) =itObs->range;
    	Z(row,1) =itObs->yaw;

    	ASSERTMSG_(itObs->pitch==0,"ERROR: Observation contains pitch!=0 but this is SLAM in 2D only!!!")

    	// Fill data asociation:
    	//  Using IDs!
    	std::map<CLandmark::TLandmarkID,unsigned int>::iterator itID;
    	if ( (itID = m_IDs.find( itObs->landmarkID )) != m_IDs.end()  )
    		*itDA = itID->second; // This row in Z corresponds to the i'th map element in the state vector:
    }

	// R2 vector:
	// ---------------------
    R2.resize(2);
    R2[0] = square( options.std_sensor_range );
    R2[1] = square( options.std_sensor_yaw );

	MRPT_TRY_END

}


/** This virtual function musts normalize the state vector and covariance matrix (only if its necessary).
 */
void  CRangeBearingKFSLAM2D::OnNormalizeStateVector()
{
	MRPT_TRY_START

	// Check angles:
	mrpt::math::wrapToPiInPlace(m_xkk[2]);

	MRPT_TRY_END
}

/*---------------------------------------------------------------
						loadOptions
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM2D::loadOptions( const mrpt::utils::CConfigFileBase &ini )
{
	// Main
	options.loadFromConfigFile( ini, "RangeBearingKFSLAM" );
	KF_options.loadFromConfigFile( ini, "RangeBearingKFSLAM_KalmanFilter" );
}

/*---------------------------------------------------------------
						loadOptions
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM2D::TOptions::loadFromConfigFile(
	const mrpt::utils::CConfigFileBase	&source,
	const std::string		&section)
{
	stds_Q_no_odo[2] = RAD2DEG(stds_Q_no_odo[2]);

	source.read_vector(section,"stds_Q_no_odo", stds_Q_no_odo, stds_Q_no_odo );
	ASSERT_(stds_Q_no_odo.size()==3)

	stds_Q_no_odo[2] = DEG2RAD(stds_Q_no_odo[2]);

	std_sensor_range	= source.read_float(section,"std_sensor_range", std_sensor_range);
	std_sensor_yaw		= DEG2RAD( source.read_float(section,"std_sensor_yaw_deg", RAD2DEG(std_sensor_yaw)));

	MRPT_LOAD_CONFIG_VAR(quantiles_3D_representation, float, source,section);
	MRPT_LOAD_CONFIG_VAR(create_simplemap, bool, source,section)
}

/*---------------------------------------------------------------
						Constructor
  ---------------------------------------------------------------*/
CRangeBearingKFSLAM2D::TOptions::TOptions() :
	stds_Q_no_odo(3,0),
	std_sensor_range ( 0.1f),
	std_sensor_yaw   ( DEG2RAD( 0.5f )),
	quantiles_3D_representation ( 3),
	create_simplemap			(false)
{
	stds_Q_no_odo[0]=0.10;
	stds_Q_no_odo[1]=0.10;
	stds_Q_no_odo[2]=DEG2RAD(4.0f);
}

/*---------------------------------------------------------------
						dumpToTextStream
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM2D::TOptions::dumpToTextStream( CStream &out)const
{
	out.printf("\n----------- [CRangeBearingKFSLAM2D::TOptions] ------------ \n\n");

	//out.printf("stds_Q_no_odo                           = "); stds_Q_no_odo

}

/** If applicable to the given problem, this method implements the inverse observation model needed to extend the "map" with a new "element".
* \param in_z This is the same matrix returned by OnGetObservations().
* \param in_obsIndex The index of the observation whose inverse sensor is to be computed. It corresponds to the row in in_z where the observation can be found.
* \param in_idxNewFeat The index that this new feature will have in the state vector (0:just after the vehicle state, 1: after that,...). Save this number so data association can be done according to these indices.
* \param out_yn The F-length vector with the inverse observation model \f$ y_n=y(x,z_n) \f$.
* \param out_dyn_dxv The \f$F \times V\f$ Jacobian of the inv. sensor model wrt the robot pose \f$ \frac{\partial y_n}{\partial x_v} \f$.
* \param out_dyn_dhn The \f$F \times O\f$ Jacobian of the inv. sensor model wrt the observation vector \f$ \frac{\partial y_n}{\partial h_n} \f$.
*
*  - O: get_observation_size()
*  - V: get_vehicle_size()
*  - F: get_feature_size()
*/
void  CRangeBearingKFSLAM2D::OnInverseObservationModel(
	const CMatrixTemplateNumeric<KFTYPE> &in_z,
	const size_t                         &in_obsIdx,
	const size_t                         &in_idxNewFeat,
	CVectorTemplate<KFTYPE>              &yn,
	CMatrixTemplateNumeric<KFTYPE>       &dyn_dxv,
	CMatrixTemplateNumeric<KFTYPE>       &dyn_dhn )
{
	MRPT_TRY_START

	/* -------------------------------------------
       Equations, obtained using matlab

        x0 y0 phi0      % Robot's 2D pose
        x0s y0s phis    % Sensor's 2D pose relative to robot
        hr ha         	% Observation hn: range, yaw

        xi yi           % Absolute 2D landmark coordinates:

        dyn_dxv   -> Jacobian of the inv. observation model wrt the robot pose
        dyn_dhn   -> Jacobian of the inv. observation model wrt each landmark observation

        Sizes:
	     hn:      1x2  <--
         yn:      1x2  -->
         dyn_dxv: 2x3  -->
         dyn_dhn: 2x2  -->
	  ------------------------------------------- */

	CObservationBearingRangePtr obs = m_SF->getObservationByClass<CObservationBearingRange>();
	ASSERT_(obs);
	const CPose2D   sensorPoseOnRobot = CPose2D( obs->sensorLocationOnRobot );

	// Mean of the prior of the robot pose:
	const TPose2D  robotPose( m_xkk[0],m_xkk[1],m_xkk[2] );

	yn.resize(2);
    dyn_dxv.setSize(2,3);
    dyn_dhn.setSize(2,2);

    // Robot 2D pose:
    const KFTYPE x0 = m_xkk[0];
    const KFTYPE y0 = m_xkk[1];
    const KFTYPE phi0  = m_xkk[2];

    const KFTYPE cphi0 = cos(phi0);
    const KFTYPE sphi0 = sin(phi0);

    // Sensor 6D pose on robot:
    const KFTYPE x0s = sensorPoseOnRobot.x();
    const KFTYPE y0s = sensorPoseOnRobot.y();
    const KFTYPE phis = sensorPoseOnRobot.phi();

    ASSERT_(in_z.getColCount()==2)
    const KFTYPE hr = in_z(in_obsIdx,0);
    const KFTYPE ha = in_z(in_obsIdx,1);

    const KFTYPE cphi_0sa = cos(phi0+phis+ha);
    const KFTYPE sphi_0sa = sin(phi0+phis+ha);

    // Compute the mean 2D absolute coordinates:
    yn[0] = hr*cphi_0sa+cphi0*x0s-sphi0*y0s+x0;
    yn[1] = hr*sphi_0sa+sphi0*x0s+cphi0*y0s+y0;

    // Jacobian wrt xv:
    dyn_dxv(0,0) = 1;
    dyn_dxv(0,1) = 0;
    dyn_dxv(0,2) = -hr*sphi_0sa-sphi0*x0s-cphi0*y0s;

    dyn_dxv(1,0) = 0;
    dyn_dxv(1,1) = 1;
    dyn_dxv(1,2) = hr*cphi_0sa+cphi0*x0s-sphi0*y0s;

    // Jacobian wrt hn:
    dyn_dhn(0,0) = cphi_0sa;
    dyn_dhn(0,1) = -hr*sphi_0sa;

    dyn_dhn(1,0) = sphi_0sa;
    dyn_dhn(1,1) = hr*cphi_0sa;

	// ----------------------------------------------
	// introduce in the lists of ID<->index in map:
	// ----------------------------------------------
	ASSERT_( in_obsIdx < obs->sensedData.size() )
	ASSERT_( obs->sensedData[in_obsIdx].landmarkID>=0 )

	m_IDs[ obs->sensedData[in_obsIdx].landmarkID ] = in_idxNewFeat;
	m_IDs_inverse[ in_idxNewFeat ] = obs->sensedData[in_obsIdx].landmarkID;


    MRPT_TRY_END
}


/*---------------------------------------------------------------
						getAs3DObject
  ---------------------------------------------------------------*/
void  CRangeBearingKFSLAM2D::getAs3DObject( mrpt::opengl::CSetOfObjectsPtr	&outObj ) const
{
    outObj->clear();

	// ------------------------------------------------
	//  Add the XYZ corner for the current area:
	// ------------------------------------------------
	outObj->insert( opengl::stock_objects::CornerXYZ() );


	// 2D ellipsoid for robot pose:
	CPoint2DPDFGaussian pointGauss;
    pointGauss.mean.x( m_xkk[0] );
    pointGauss.mean.y( m_xkk[1] );
    CMatrixTemplateNumeric<KFTYPE>  COV;
    m_pkk.extractMatrix(0,0,2,2, COV);
    pointGauss.cov = COV;

    {
		opengl::CEllipsoidPtr ellip = opengl::CEllipsoid::Create();

		ellip->setPose(pointGauss.mean);
		ellip->setCovMatrix(pointGauss.cov);
		ellip->enableDrawSolid3D(false);
		ellip->setQuantiles( options.quantiles_3D_representation );
		ellip->set3DsegmentsCount(10);
		ellip->setColor(1,0,0);

		outObj->insert( ellip );
    }


	// 2D ellipsoids for landmarks:
	size_t i,nLMs = (m_xkk.size()-3)/2;
	for (i=0;i<nLMs;i++)
	{
        pointGauss.mean.x( m_xkk[3+2*i+0] );
        pointGauss.mean.y( m_xkk[3+2*i+1] );
        m_pkk.extractMatrix(3+2*i,3+2*i,2,2, COV);
        pointGauss.cov = COV;

		opengl::CEllipsoidPtr ellip = opengl::CEllipsoid::Create();

		ellip->setName( format( "%"PRIuPTR,i ) );
		ellip->enableShowName(true);
		ellip->setPose( pointGauss.mean );
		ellip->setCovMatrix( pointGauss.cov );
		ellip->enableDrawSolid3D(false);
		ellip->setQuantiles( options.quantiles_3D_representation );
		ellip->set3DsegmentsCount(10);

		ellip->setColor(0,0,1);

		outObj->insert( ellip );
	}
}



/*---------------------------------------------------------------
              saveMapAndPathRepresentationAsMATLABFile
  ---------------------------------------------------------------*/
void CRangeBearingKFSLAM2D::saveMapAndPath2DRepresentationAsMATLABFile(
	const string &fil,
	float stdCount,
	const string &styleLandmarks,
	const string &stylePath,
	const string &styleRobot ) const
{
	FILE	*f= os::fopen(fil.c_str(),"wt");
	if (!f) return;

	CMatrixDouble cov(2,2);
	CVectorDouble mean(2);

	// Header:
	os::fprintf(f,"%%--------------------------------------------------------------------\n");
	os::fprintf(f,"%% File automatically generated using the MRPT method:\n");
	os::fprintf(f,"%% 'CRangeBearingKFSLAM2D::saveMapAndPath2DRepresentationAsMATLABFile'\n");
	os::fprintf(f,"%%\n");
	os::fprintf(f,"%%                        ~ MRPT ~\n");
	os::fprintf(f,"%%  Jose Luis Blanco Claraco, University of Malaga @ 2008\n");
	os::fprintf(f,"%%  http://mrpt.sourceforge.net/ \n");
	os::fprintf(f,"%%--------------------------------------------------------------------\n");

	// Main code:
	os::fprintf(f,"hold on;\n\n");

	size_t i, nLMs = (m_xkk.size()-get_vehicle_size())/get_feature_size();

	for (i=0;i<nLMs;i++)
	{
		size_t idx = get_vehicle_size()+i*get_feature_size();

		cov(0,0) = m_pkk(idx+0,idx+0);
		cov(1,1) = m_pkk(idx+1,idx+1);
		cov(0,1) = cov(1,0) = m_pkk(idx+0,idx+1);

		mean[0] = m_xkk[idx+0];
		mean[1] = m_xkk[idx+1];

		// Command to draw the 2D ellipse:
		os::fprintf(f, "%s", math::MATLAB_plotCovariance2D( cov, mean, stdCount,styleLandmarks ).c_str() );
	}


	// Now: the robot path:
	// ------------------------------
	if (m_SFs.size())
	{
		os::fprintf(f,"\nROB_PATH=[");
		for (i=0;i<m_SFs.size();i++)
		{
			CSensoryFramePtr dummySF;
			CPose3DPDFPtr pdf3D;
			m_SFs.get(i,pdf3D,dummySF);

			CPose3D p;
			pdf3D->getMean(p);
			CPoint3D pnt3D(p); // 6D -> 3D only

			os::fprintf(f,"%.04f %.04f", pnt3D.x(), pnt3D.y() );
			if (i<(m_SFs.size()-1))
				os::fprintf(f,";");
		}
		os::fprintf(f,"];\n");

		os::fprintf(f,"plot(ROB_PATH(:,1),ROB_PATH(:,2),'%s');\n",stylePath.c_str());
	}

	// The robot pose:
	cov(0,0) = m_pkk(0,0);
	cov(1,1) = m_pkk(1,1);
	cov(0,1) = cov(1,0) = m_pkk(0,1);

	mean[0] = m_xkk[0];
	mean[1] = m_xkk[1];

	os::fprintf(f, "%s", math::MATLAB_plotCovariance2D( cov, mean, stdCount, styleRobot ).c_str() );

	os::fprintf(f,"\naxis equal;\n");
	os::fclose(f);
}
