/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Machine Perception and Intelligent    |
   |      Robotics Lab, University of Malaga (Spain).                          |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/precomp_core.h>  // Only for precomp. headers, include all libmrpt-core headers.

#include <mrpt/slam/data_association.h>
#include <mrpt/math/distributions.h>  // for chi2inv
#include <mrpt/poses/CPointPDFGaussian.h>

#include <numeric>  // accumulate


#define DAT_ASOC_USE_KDTREE


#ifdef DAT_ASOC_USE_KDTREE
	#include <mrpt/otherlibs/ann/ANN.h>   // ANN: for kd-tree
#endif


using namespace std;
using namespace mrpt;
using namespace mrpt::math;
using namespace mrpt::poses;
using namespace mrpt::slam;
using namespace mrpt::utils;


namespace mrpt
{
	namespace slam
	{
		double joint_mahalanobis2 (
			const std::map<size_t, mrpt::poses::CPointPDFGaussian>  & predictions,
			const std::map<size_t, mrpt::poses::CPointPDFGaussian>  & observations,
			const std::map<size_t,size_t>	&currentAssociation )
		{
			return 1;
		}

		/* Based on MATLAB code by:
		  University of Zaragoza
		  Centro Politecnico Superior
		  Robotics and Real Time Group
		  Authors:  J. Neira, J. Tardos
		*/
		void JCBB_recursive(
			const std::map<size_t, mrpt::poses::CPointPDFGaussian>  & predictions,
			const std::map<size_t, mrpt::poses::CPointPDFGaussian>  & observations,
			TDataAssociationResults			&results,
			const std::map<size_t,size_t>	&currentAssociation,
			const std::map<size_t, mrpt::poses::CPointPDFGaussian>::const_iterator &itObservation )
		{
			// End of iteration?
			if (itObservation==observations.end())
			{
				if (currentAssociation.size()>results.associations.size())
				{
					// It's a better choice:
					results.associations = currentAssociation;
					results.distance_maha = joint_mahalanobis2 (predictions, observations, currentAssociation);
				}
				else if ( currentAssociation.size()==results.associations.size() )
				{
					const double d2 = joint_mahalanobis2 (predictions, observations, currentAssociation);
					if (d2 < results.distance_maha)
					{
						results.associations = currentAssociation;
						results.distance_maha = d2;
					}
				}
			}
			else // A normal iteration:
			{
				// Iterate for all compatible landmarsk of "itObservation"
				const size_t obsIdx = results.id2obsidx[ itObservation->first ];
				const size_t nPreds = results.indiv_compatibility.getRowCount();
				for (size_t predIdx=0;predIdx<nPreds;predIdx++)
				{
					if ( results.indiv_compatibility(predIdx,obsIdx) )
					{
						const size_t predID = results.predidx2id[predIdx];
						// Only if predID is NOT already assigned:
						bool already_asigned = false;
						for (map<size_t,size_t>::const_iterator itS=currentAssociation.begin();itS!=currentAssociation.end();++itS)
						{
							if (itS->second==predID)
							{
								already_asigned = true;
								break;
							}
						}
						if (!already_asigned)
						{
							map<size_t,size_t>	new_assoc = currentAssociation;
							new_assoc[ itObservation->first ] = results.predidx2id[predIdx];

							map<size_t, CPointPDFGaussian>::const_iterator new_obs = itObservation;
							++new_obs;

							JCBB_recursive(predictions,observations, results, new_assoc, new_obs);
						}
					}
				}

				// Can we do it better?
				// potentials  = pairings(compatibility.AL(i+1:end))
				const size_t potentials = std::accumulate( results.indiv_compatibility_counts.begin()+(obsIdx+1), results.indiv_compatibility_counts.end(),0 );
				if ((currentAssociation.size() + potentials) >= results.associations.size() )
				{
					// star node: Ei not paired
					map<size_t, CPointPDFGaussian>::const_iterator new_obs = itObservation;
					++new_obs;

					JCBB_recursive(predictions,observations, results, currentAssociation, new_obs);
				}
			}
		}
	}
}

/*---------------------------------------------------------------
					data_association
  ---------------------------------------------------------------*/
void mrpt::slam::data_association(
	const std::map<size_t, mrpt::poses::CPointPDFGaussian>  & predictions,
	const std::map<size_t, mrpt::poses::CPointPDFGaussian>  & observations,
	TDataAssociationResults   &results,
	const TDataAssociationMethod  method,
	const double  chi2quantile
	)
{
	MRPT_TRY_START

	ASSERT_( !predictions.empty() )
	ASSERT_( !observations.empty() )
	ASSERT_( chi2quantile>0 && chi2quantile<1 )

	const double chi2thres = mrpt::math::chi2inv( chi2quantile, 3 ); // #dims = 3


	std::map<size_t, mrpt::poses::CPointPDFGaussian>::const_iterator itPred;
	std::map<size_t, mrpt::poses::CPointPDFGaussian>::const_iterator itObs;
	size_t i,j;

	// ------------------------------------------------------------
	// Build a KD-tree of the predictions for quick look-up:
	// ------------------------------------------------------------
	const size_t  	nPredictions  = predictions.size();
	const size_t  	nObservations = observations.size();

#ifdef DAT_ASOC_USE_KDTREE
	const size_t N_KD_RESULTS = nPredictions; //std::min(size_t(10),nPredictions);
	std::vector<ANNdist>	kd_result_distances(N_KD_RESULTS);
	std::vector<ANNidx>		kd_result_indices(N_KD_RESULTS);
	ANNcoord		kd_queryPoint[3];
	ANNpointArray 	kd_dataPoints = annAllocPts(nPredictions,3);

	for(i=0,itPred=predictions.begin();i<nPredictions;i++,itPred++)
	{
		kd_dataPoints [i][0]= itPred->second.mean.x();
		kd_dataPoints [i][1]= itPred->second.mean.y();
		kd_dataPoints [i][2]= itPred->second.mean.z();
	}

	// Construct kd-tree for the predictions:
	ANNkd_tree	kd_tree(kd_dataPoints,nPredictions,3);
#endif

	results.distance_maha = std::numeric_limits<double>::max();

	// Build the bi-maps:
	results.obsidx2id.clear();	results.predidx2id.clear();
	results.id2obsidx.clear();	results.id2predidx.clear();
	for (i=0,itPred=predictions.begin();i<nPredictions;i++,itPred++)
	{
		results.predidx2id[i] = itPred->first;
		results.id2predidx[itPred->first] = i;
	}
	for (j=0,itObs=observations.begin();j<nObservations;j++,itObs++)
	{
		results.obsidx2id[j]=itObs->first;
		results.id2obsidx[itObs->first] = j;
	}


	//-------------------------------------------
	// Compute the individual compatibility:
	//-------------------------------------------
	results.indiv_distances.resize(nPredictions,nObservations);
	results.indiv_compatibility.setSize(nPredictions,nObservations);
	results.indiv_compatibility_counts.assign(nObservations,0);

	results.indiv_distances.fill( 1000 );
	results.indiv_compatibility.fill(false);

	for (j=0,itObs=observations.begin();j<nObservations;j++,itObs++)
	{
#ifndef DAT_ASOC_USE_KDTREE
		// Compute all the distances:
		for (i=0,itPred=predictions.begin();i<nPredictions;i++,itPred++)
		{
			double d2 = square( itPred->second.mahalanobisDistanceTo( itObs->second ) );
			results.indiv_distances(i,j) = d2;
			if (true == (results.indiv_compatibility(i,j) = d2 < chi2thres ) )
				results.indiv_compatibility_counts[j]++;
		}
#else
		// Use a kd-tree and compute only the N closest ones:
		kd_queryPoint[0] = itObs->second.mean.x();
		kd_queryPoint[1] = itObs->second.mean.y();
		kd_queryPoint[2] = itObs->second.mean.z();

		kd_tree.annkSearch(kd_queryPoint, N_KD_RESULTS, &kd_result_indices[0], &kd_result_distances[0] );

		// Only compute the distances for these ones:
		for (size_t w=0;w<N_KD_RESULTS;w++)
		{
			const size_t i = kd_result_indices[w];
			const size_t id_pred = results.predidx2id[i];
			double d2 = square( itObs->second.mahalanobisDistanceTo( predictions.find(id_pred)->second ) );
			if (d2>6*chi2thres) break; // Since kd-tree returns the landmarks by distance order, we can skip the rest

			results.indiv_distances(i,j) = d2;
			if (true==(results.indiv_compatibility(i,j) = d2 < chi2thres ))
				results.indiv_compatibility_counts[j]++;
		}
#endif
	}

#if 0
	cout << "Distances: " << endl << results.indiv_distances << endl;
	cout << "invdiv compat: " << endl << results.indiv_compatibility << endl;
#endif

	// Do associations:
	results.associations.clear();

	switch(method)
	{
		// --------------------------
		// Nearest-neighbor
		// --------------------------
	case assocNN:
		{
			for (i=0,itPred=predictions.begin();i<nPredictions;i++,itPred++)
			{
				double min_dist2 = std::numeric_limits<double>::max();
				size_t best_assoc = 0;
				for (j=0,itObs=observations.begin();j<nObservations;j++,itObs++)
				{
					const double d2 = results.indiv_distances.get_unsafe(i,j);
					if (d2<min_dist2)
					{
						min_dist2 = d2;
						best_assoc = j;
					}
				}
				results.associations[i] = best_assoc;
			}
		}
		break;

		// ------------------------------------
		// Joint Compatibility Branch & Bound:
		// ------------------------------------
	case assocJCBB:
		{
			// Call to the recursive method:
			std::map<size_t,size_t>  currentAssociation;
			JCBB_recursive(predictions, observations, results, currentAssociation, observations.begin() );
		}
		break;

	default:
		THROW_EXCEPTION("Unknown value of 'method'")
	};

#ifdef DAT_ASOC_USE_KDTREE
	annDeallocPts(kd_dataPoints);	// free temporary points for kd-tree
#endif

	MRPT_TRY_END
}

