/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Machine Perception and Intelligent    |
   |      Robotics Lab, University of Malaga (Spain).                          |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/precomp_core.h>  // Only for precomp. headers, include all libmrpt-core headers.

#include <mrpt/vision/CFeatureExtraction.h>

using namespace mrpt;
using namespace mrpt::vision;
using namespace mrpt::system;
using namespace std;


/************************************************************************************************
*								Constructor	        									        *
************************************************************************************************/
CFeatureExtraction::CFeatureExtraction()
{
}

/************************************************************************************************
*								Destructor          									        *
************************************************************************************************/
CFeatureExtraction::~CFeatureExtraction()
{
}

/************************************************************************************************
*								extractFeatures  									        *
************************************************************************************************/
void  CFeatureExtraction::detectFeatures(
		const CImage		&img,
		CFeatureList			&feats,
		unsigned int			init_ID,
		unsigned int			nDesiredFeatures,
		const TImageROI			&ROI) const
{
	switch( options.featsType )
	{
		case featHarris:
			extractFeaturesKLT(img, feats, init_ID, nDesiredFeatures, ROI);
			break;

		case featKLT:
			extractFeaturesKLT(img, feats, init_ID, nDesiredFeatures, ROI);
			break;

		case featSIFT:
			extractFeaturesSIFT(img, feats, init_ID, nDesiredFeatures, ROI);
			break;

		case featBCD:
			extractFeaturesBCD(img, feats, init_ID, nDesiredFeatures, ROI);
			break;

		case featSURF:
			extractFeaturesSURF(img, feats, init_ID, nDesiredFeatures, ROI);
			break;

		default:
			THROW_EXCEPTION("options.method has an invalid value!");
			break;
	}
}

/************************************************************************************************
 								computeDescriptors
************************************************************************************************/
void  CFeatureExtraction::computeDescriptors(
	const CImage	&in_img,
	CFeatureList		&inout_features,
	TDescriptorType		in_descriptor_list) const
{
	MRPT_TRY_START

	int nDescComputed = 0;

	if ((in_descriptor_list & descSIFT) != 0)
	{
		this->internal_computeSiftDescriptors(in_img,inout_features);
		++nDescComputed;
	}
	if ((in_descriptor_list & descSURF) != 0)
	{
		this->internal_computeSurfDescriptors(in_img,inout_features);
		++nDescComputed;
	}
	if ((in_descriptor_list & descSpinImages) != 0)
	{
		this->internal_computeSpinImageDescriptors(in_img,inout_features);
		++nDescComputed;
	}
	if ((in_descriptor_list & descPolarImages) != 0)
	{
		this->internal_computePolarImageDescriptors(in_img,inout_features);
		++nDescComputed;
	}
	if ((in_descriptor_list & descLogPolarImages) != 0)
	{
		this->internal_computeLogPolarImageDescriptors(in_img,inout_features);
		++nDescComputed;
	}

	if (!nDescComputed)
		THROW_EXCEPTION_CUSTOM_MSG1("No known descriptor value found in in_descriptor_list=%u",(unsigned)in_descriptor_list)

	MRPT_TRY_END
}

/************************************************************************************************
*								extractFeaturesBCD  									        *
************************************************************************************************/
void  CFeatureExtraction::extractFeaturesBCD(
		const CImage		&img,
		CFeatureList			&feats,
		unsigned int			init_ID,
		unsigned int			nDesiredFeatures,
		const TImageROI			&ROI) const
{
	MRPT_UNUSED_PARAM(img);
	MRPT_UNUSED_PARAM(feats);
	MRPT_UNUSED_PARAM(nDesiredFeatures);
	MRPT_UNUSED_PARAM(ROI);

	THROW_EXCEPTION("Not implemented yet!");
} // end extractFeaturesBCD



/*------------------------------------------------------------
					TOptions()
-------------------------------------------------------------*/
CFeatureExtraction::TOptions::TOptions()
{
	// General options
	featsType		= featKLT;				// Default Method: Kanade-Lucas-Tomasi
	patchSize		= 21;					// Patch size
	FIND_SUBPIXEL	= true;					// Find subpixel

	// Harris Options
	harrisOptions.k				= 0.04f;
	harrisOptions.radius		= 3;		//15;
	harrisOptions.threshold		= 0.005f;	//0.01f; The lower this is, more features will be found
	harrisOptions.sigma			= 3.0f;
	harrisOptions.min_distance	= 5;		//10;

	// KLT Options
	KLTOptions.min_distance		= 5;		//10;
	KLTOptions.threshold		= 0.005f;	//0.01f;
	KLTOptions.radius			= 3;		//15;

	// SIFT Options
	SIFTOptions.implementation	= Hess;		// Default implementation: Hess

	// SURF Options
	SURFOptions.rotation_invariant	= true;

	// BCD Options

	// SpinImages Options:
	SpinImagesOptions.hist_size_distance  = 10;
	SpinImagesOptions.hist_size_intensity = 10;
	SpinImagesOptions.radius              = 20;
	SpinImagesOptions.std_dist            = 0.4f;
	SpinImagesOptions.std_intensity       = 10;

	// TPolarImagesOptions
	PolarImagesOptions.bins_angle		= 8;
	PolarImagesOptions.bins_distance	= 6;
	PolarImagesOptions.radius			= 20;

	// LogPolarImagesOptions
	LogPolarImagesOptions.radius		= 30;
	LogPolarImagesOptions.num_angles	= 16; // Log-Polar image patch will have dimensions WxH, with:  W=num_angles,  H= rho_scale * log(radius)
	LogPolarImagesOptions.rho_scale		= 5;

}

/*---------------------------------------------------------------
					dumpToTextStream
  ---------------------------------------------------------------*/
void CFeatureExtraction::TOptions::dumpToTextStream(CStream	&out) const
{
	out.printf("\n----------- [CFeatureExtraction::TOptions] ------------ \n\n");

	LOADABLEOPTS_DUMP_VAR(featsType,int)
	LOADABLEOPTS_DUMP_VAR(patchSize, int)
	LOADABLEOPTS_DUMP_VAR(FIND_SUBPIXEL, bool)

	LOADABLEOPTS_DUMP_VAR(harrisOptions.k,double)
	LOADABLEOPTS_DUMP_VAR(harrisOptions.radius,int)
	LOADABLEOPTS_DUMP_VAR(harrisOptions.threshold,float)
	LOADABLEOPTS_DUMP_VAR(harrisOptions.sigma,float)
	LOADABLEOPTS_DUMP_VAR(harrisOptions.min_distance,float)

	LOADABLEOPTS_DUMP_VAR(KLTOptions.min_distance,float)
	LOADABLEOPTS_DUMP_VAR(KLTOptions.threshold,float)
	LOADABLEOPTS_DUMP_VAR(KLTOptions.radius,int)

	LOADABLEOPTS_DUMP_VAR(SIFTOptions.implementation,int)

	LOADABLEOPTS_DUMP_VAR(SURFOptions.rotation_invariant,bool)

	LOADABLEOPTS_DUMP_VAR(SpinImagesOptions.hist_size_distance,int)
	LOADABLEOPTS_DUMP_VAR(SpinImagesOptions.hist_size_intensity,int)
	LOADABLEOPTS_DUMP_VAR(SpinImagesOptions.radius,int)
	LOADABLEOPTS_DUMP_VAR(SpinImagesOptions.std_dist,float)
	LOADABLEOPTS_DUMP_VAR(SpinImagesOptions.std_intensity,float)

	LOADABLEOPTS_DUMP_VAR(PolarImagesOptions.bins_angle,int)
	LOADABLEOPTS_DUMP_VAR(PolarImagesOptions.bins_distance,int)
	LOADABLEOPTS_DUMP_VAR(PolarImagesOptions.radius,int)

	LOADABLEOPTS_DUMP_VAR(LogPolarImagesOptions.radius,int)
	LOADABLEOPTS_DUMP_VAR(LogPolarImagesOptions.num_angles,int)
	LOADABLEOPTS_DUMP_VAR(LogPolarImagesOptions.rho_scale,double)

	out.printf("\n");
}

/*---------------------------------------------------------------
					loadFromConfigFile
  ---------------------------------------------------------------*/
void CFeatureExtraction::TOptions::loadFromConfigFile(
	const mrpt::utils::CConfigFileBase  &iniFile,
	const std::string &section)
{
	MRPT_LOAD_CONFIG_VAR_CAST(featsType, int, TFeatureType, iniFile,section)
	MRPT_LOAD_CONFIG_VAR(patchSize, int,  iniFile, section)
	MRPT_LOAD_CONFIG_VAR(FIND_SUBPIXEL, bool,  iniFile, section)

	//string sect = section;
	MRPT_LOAD_CONFIG_VAR(harrisOptions.k,double,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(harrisOptions.radius,int,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(harrisOptions.threshold,float,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(harrisOptions.sigma,float,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(harrisOptions.min_distance,float,  iniFile,section)

	MRPT_LOAD_CONFIG_VAR(KLTOptions.min_distance,float,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(KLTOptions.threshold,float,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(KLTOptions.radius,int,  iniFile,section)

	MRPT_LOAD_CONFIG_VAR_CAST(SIFTOptions.implementation,int,TSIFTImplementation, iniFile,section)

	MRPT_LOAD_CONFIG_VAR(SURFOptions.rotation_invariant,bool,  iniFile,section)

	MRPT_LOAD_CONFIG_VAR(SpinImagesOptions.hist_size_distance,int,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(SpinImagesOptions.hist_size_intensity,int,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(SpinImagesOptions.radius,int,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(SpinImagesOptions.std_dist,float,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(SpinImagesOptions.std_intensity,float,  iniFile,section)

	MRPT_LOAD_CONFIG_VAR(PolarImagesOptions.bins_angle,int,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(PolarImagesOptions.bins_distance,int,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(PolarImagesOptions.radius,int,  iniFile,section)

	MRPT_LOAD_CONFIG_VAR(LogPolarImagesOptions.radius,int,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(LogPolarImagesOptions.num_angles,int,  iniFile,section)
	MRPT_LOAD_CONFIG_VAR(LogPolarImagesOptions.rho_scale,double,  iniFile,section)

}

