/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Perception and Robotics               |
   |      research group, University of Malaga (Spain).                        |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */


/**************************************************************************
 Application: mono-slam

  - 2007:
	The first version of this application, the corresponding classes
	and config files were developed by Antonio J. Ortiz de Galisteo
	as part of his Master Thesis at the University of Malaga (2007).

  - Jan-Feb 2009:
	mono-slam updated and integrated back into MRPT by M.A. Amiri Atashgah (MAAA)
	at the University of Sharif Technology, Tehran, Iran (2009).

 **************************************************************************/

#include <mrpt/monoslam/CMonoSlamMotionModel.h>
#include <mrpt/core.h>

using namespace mrpt;
using namespace mrpt::utils;
using namespace mrpt::poses;
using namespace mrpt::slam;
using namespace mrpt::math;
using namespace mrpt::system;
using namespace mrpt::monoslam;

/*---------------------------------------------------------------
					Constructor
 ---------------------------------------------------------------*/
CMonoSlamMotionModel::CMonoSlamMotionModel() :
	sigma_ang_vel	(6.0),  // rad/s^2
	sigma_lin_vel		(6.0),  // m/s^2
	delta_t				(0.06333)  // 15Hz
{

}

//Here we can use the specific dimension because is a specific class for monocular SLAM with
//Inverse Depth Parametrization
/************************************************************************************************************/
void CMonoSlamMotionModel::extract_components(CVectorDouble &xv,
                                              CVectorDouble &r,
											  CQuaternionDouble &q,
											  CVectorDouble &v,
											  CVectorDouble &w)
{
	if (xv.size()!= state_size)
		THROW_EXCEPTION("extract_components MonoSlamMotionModel: Wrong dimension in State Vector");

	r.resize(3,0);	v.resize(3,0);	w.resize(3,0);

	xv.extract_vector(0,r);
	xv.extract_vector(3,q);
	xv.extract_vector(7,v);
	xv.extract_vector(10,w);
}

/************************************************************************************************************/
void CMonoSlamMotionModel::compose_xv(CVectorDouble &xv,
                                      const CVectorDouble &r,
									  const CQuaternionDouble &q,
									  const CVectorDouble &v,
									  const CVectorDouble &w)
{
	if ((r.size()!= 3)||(q.size()!=4)||(v.size()!= 3)||(w.size()!= 3))
		THROW_EXCEPTION("compose_xv MonoSlamMotionModel: Wrong dimension in State Vector Components");

	xv.insert_vector(0,r);
	xv.insert_vector(3,q);
	xv.insert_vector(7,v);
	xv.insert_vector(10,w);
}

/**************************************************************************************************************/
/**	State Update (for prediction)
*/

void CMonoSlamMotionModel::func_fv(CVectorDouble &xv)
{

// r_{new} = r + (v+V)*Delta t
// q_{new} = q * q(omega + Omega) * Delta t)
// v_{new} = v + V
// w_{new} = w + W

//[x y z qr qx qy qz vx vy vz wx wy wz]
// 0 1 2 3  4  5  6  7  8  9  10 11 12
	CVectorDouble r,v,w,vdelta;	//position, vel_lin, vel_ang
	CQuaternionDouble qold,qnew;
    extract_components(xv,r,qold,v,w);

	vdelta = v * delta_t;		// (Mod: JLBC: In this way a warning is avoided 19-DEC-2006)
	r = r + vdelta;

	CQuaternionDouble q_temp(w*delta_t);
	qnew.q_prod(qold,q_temp);

	compose_xv(xv,r,qnew,v,w);

	//v y w no se modifican si no usamos aceleracion (en matlab no se modifica)
}

/************************************************************************************************************************/

void CMonoSlamMotionModel::func_dfv_by_dxv(CVectorDouble &xv, CMatrixDouble &F)
{
	// Now on to the Jacobian...
	// Identity is a good place to start since overall structure is like this
	// I       0             I * delta_t   0
	// 0       dqnew_by_dq   0             dqnew_by_domega
	// 0       0             I             0
	// 0       0             0             I

	ASSERT_(xv.size()==state_size);
	F.setSize(state_size,state_size);
	F.unit();

	// Fill in F = I * delta_t
    CMatrixDouble Temp33A(3,3);
	for (int i=0;i<3;i++)
        Temp33A(i,i)=delta_t;

	F.insertMatrix(0,7,Temp33A);	// (0,7) is the place for the "I * delta_t" 3x3 submatrix

    // Fill in dqnew_by_dq
	// Warning: This function must use x(k|k), not x(k+1|k) --> Beware of not calling it after "func_fv", which modified "x_k_k"
	CVectorDouble w_old(3); 
	CQuaternionDouble q_old;
	xv.extract_vector(10,w_old);
	xv.extract_vector(3,q_old);
	CQuaternionDouble qwt(w_old*delta_t);

	CMatrixDouble Temp44A(4,4);
	dq3_dq2(qwt,Temp44A);

	F.insertMatrix(3,3,Temp44A);

	// Fill in dqnew_by_domega = d(q x qwt)_by_dqwt . dqwt_by_domega
	dq3_dq1(q_old,Temp44A); // Temp44A is d(q x qwt) by dqwt

    // Use function below for dqwt_by_domega
	CMatrixDouble Temp43A(4,3);
    dqomegadt_by_domega(w_old, Temp43A);

	// Multiply them together
    CMatrixDouble Temp43B(4,3);
	Temp43B	= Temp44A * Temp43A;
    // And plug it in
	F.insertMatrix(3,10,Temp43B);
}

/************************************************************************************************************/
void CMonoSlamMotionModel::func_Q(CVectorDouble &xv,CMatrixDouble &QRES)
{
  // Fill noise covariance matrix Pnn: this is the covariance of  the noise vector [V Omega]^T
  // that gets added to the state.
  // Form of this could change later, but for now assume that
  // V and Omega are independent, and that each of their components is
  // independent...
  QRES.setSize(state_size,state_size);

  const double linear_velocity_noise_variance = square( sigma_lin_vel * delta_t );
  const double angular_velocity_noise_variance = square( sigma_ang_vel * delta_t );

  // Independence means that the matrix is diagonal
  CMatrixDouble Pnn(feature_size,feature_size);
  Pnn(0,0) = linear_velocity_noise_variance;
  Pnn(1,1) = linear_velocity_noise_variance;
  Pnn(2,2) = linear_velocity_noise_variance;
  Pnn(3,3) = angular_velocity_noise_variance;
  Pnn(4,4) = angular_velocity_noise_variance;
  Pnn(5,5) = angular_velocity_noise_variance;

  // Form Jacobian dxnew_by_dn
  // Is like this:
  // I * delta_t     0
  // 0               dqnew_by_dOmega
  // I               0
  // 0               I

  // Start by zeroing
  CMatrixDouble dxnew_by_dn(state_size,feature_size);

  // Fill in easy bits first
  CMatrixDouble Temp33A(3,3);
  Temp33A.unit();

  dxnew_by_dn.insertMatrix(7,0,Temp33A);
  dxnew_by_dn.insertMatrix(10,3,Temp33A);
  Temp33A *= delta_t;
  dxnew_by_dn.insertMatrix(0,0,Temp33A);

  // Tricky bit is dqnew_by_dOmega
  // Is actually the same calculation as in func_fv...
  // Since omega and Omega are additive...?

  CVectorDouble rold, vold, omegaold;
  CQuaternionDouble qold;
  extract_components(xv,rold,qold,vold,omegaold);	 // overkill but easy
  // Fill in dqnew_by_domega = d(q x qwt)_by_dqwt . dqwt_by_domega
  // Temp44A is d(q x qwt) by dqwt
  CMatrixDouble Temp44A(4,4);
  dq3_dq1(qold,Temp44A);
  // Use function below for dqwt_by_domega
  CMatrixDouble Temp43A(4,3);
  dqomegadt_by_domega(omegaold, Temp43A);
  // Multiply them together
  CMatrixDouble Temp43B(4,3);
  Temp43B = Temp44A * Temp43A;
  // And then plug into Jacobian
  dxnew_by_dn.insertMatrix(3,3,Temp43B);

  // Finally do Q = dxnew_by_dn . Pnn . dxnew_by_dnT
  QRES = dxnew_by_dn * Pnn * ~dxnew_by_dn;
}

/************************************************************************************************************/
CMatrixDouble  CMonoSlamMotionModel::XYZTPL2XYZ(CVectorDouble &xkk)
{
	//inputs:
	// xc,yc,zc : camera position in world coord system
    // theta,phi,lambda : fteature orientation and depth params

	CVectorDouble rw,mi;
	rw.resize(3);  // camera pos
	mi.resize(3);  // feature pos relative to the camera

	double theta,phi,lambda;
	unsigned int numfeatures;
	numfeatures = ((unsigned int)xkk.size() - state_size) / feature_size;
	CMatrixDouble XYZ(numfeatures,3);

	for (unsigned int i=0;i<numfeatures;i++)
	{
		xkk.extract_vector(state_size+i*feature_size,rw);
		theta=xkk[state_size+i*feature_size+3];
		phi=xkk[state_size+i*feature_size+4];
		lambda=xkk[state_size+i*feature_size+5];

		mi[0]= cos(phi) * sin(theta) / lambda;
		mi[1]= -sin(phi) / lambda;
		mi[2]= cos(phi) * cos(theta) / lambda;
		XYZ(i,0)= rw[0] + mi[0];
		XYZ(i,1)= rw[1] + mi[1];
		XYZ(i,2)= rw[2] + mi[2];
	}

 	//std::cout<<"[Mono-Slam] XYZ: \n "<< XYZ << "\n";

	return XYZ;
}

/************************************************************************************************************/
CMatrixDouble  CMonoSlamMotionModel::pxyztpl2pxyz(CMatrixDouble &pkk, CVectorDouble &xkk)
{
	unsigned int				num_feat = (unsigned int)((pkk.getRowCount() - state_size) / feature_size);
	size_t						i;
	CMatrixDouble				pxyz(3*num_feat,3*num_feat),p_aux(feature_size,feature_size),J(3,feature_size);
	CVectorDouble       		feature;
	double						theta, phi, rho, ct, st, cp, sp;
	feature.resize(6,0);

	for (i = 0 ; i < num_feat ; i++)
	{
		xkk.extract_vector(state_size+i*feature_size,feature);
		theta = feature[3];
		phi   = feature[4];
		rho   = feature[5];
		ct = cos(theta);
		st = sin(theta);
		cp = cos(phi);
		sp = sin(phi);

		pkk.extractMatrix(state_size+i*feature_size,state_size+i*feature_size,p_aux);
		
		//MAAA:Test
		//p_aux.ones();

		J(0,0) = 1.0;	J(0,3) = cp*ct/rho;		J(0,4) = -sp*st/rho;	J(0,5) = -cp*st/(rho*rho);
		J(1,1) = 1.0;							J(1,4) = -cp/rho;		J(1,5) = -sp/(rho*rho);
		J(2,2) = 1.0;	J(2,3) = -cp*st/rho;	J(2,4) = -sp*ct/rho;	J(2,5) = -cp/ct/(rho*rho);


	//std::cout<<"[Mono-Slam] pkk: \n"<< pkk << "\n";
	//std::cout<<"[Mono-Slam] J: \n "<< J << "\n";

		pxyz.insertMatrix(3*i,3*i,J * p_aux * ~J);
	}

	        
 
	return pxyz;
}
/************************************************************************************************************/
/**Calculate jacobian for func_xp. This is just the identity in this case.
*/
void CMonoSlamMotionModel::func_dxp_by_dxv(const CVectorDouble &xv, CMatrixDouble &m)
{
	MRPT_UNUSED_PARAM(xv);

	m.setSize(7,7);
	m.unit();
}

/************************************************************************************************************/
/** DQOMEGADT BY DOMEGA
* Calculate commonly used Jacobian part partial q(omega * Delta t) / partial omega
*/
void CMonoSlamMotionModel::dqomegadt_by_domega(const CVectorDouble &omega, CMatrixDouble &RESdqomegadt_by_domega)
{
  // Modulus
  double omegamod = sqrt(omega[0] * omega[0] +  omega[1] * omega[1] + omega[2] * omega[2]);
  RESdqomegadt_by_domega.setSize(4,3);

  // Use generic ancillary functions to calculate components of Jacobian
  RESdqomegadt_by_domega(0,0) = dq0_by_domegaA(omega[0], omegamod);
  RESdqomegadt_by_domega(0,1) = dq0_by_domegaA(omega[1], omegamod);
  RESdqomegadt_by_domega(0,2) = dq0_by_domegaA(omega[2], omegamod);
  RESdqomegadt_by_domega(1,0) = dqA_by_domegaA(omega[0], omegamod);
  RESdqomegadt_by_domega(1,1) = dqA_by_domegaB(omega[0], omega[1], omegamod);
  RESdqomegadt_by_domega(1,2) = dqA_by_domegaB(omega[0], omega[2], omegamod);
  RESdqomegadt_by_domega(2,0) = dqA_by_domegaB(omega[1], omega[0], omegamod);
  RESdqomegadt_by_domega(2,1) = dqA_by_domegaA(omega[1], omegamod);
  RESdqomegadt_by_domega(2,2) = dqA_by_domegaB(omega[1], omega[2], omegamod);
  RESdqomegadt_by_domega(3,0) = dqA_by_domegaB(omega[2], omega[0], omegamod);
  RESdqomegadt_by_domega(3,1) = dqA_by_domegaB(omega[2], omega[1], omegamod);
  RESdqomegadt_by_domega(3,2) = dqA_by_domegaA(omega[2], omegamod);
}



/************************************************************************************************************/
/** Ancillary functions: calculate parts of Jacobian dq_by_domega
* which are repeatable due to symmetry.
* Here omegaA is one of omegax, omegay, omegaz
* omegaB, omegaC are the other two
* And similarly with qA, qB, qC
*/

/** DQ0 BY DOMEGAA
* Ancillary function to calculate part of Jacobian partial q / partial omega which is repeatable
* due to symmetry. Here omegaA is one of omegax, omegay, omegaz.
*/
double CMonoSlamMotionModel::dq0_by_domegaA(double omegaA, double omega)
{
  return (-delta_t / 2.0f) * (omegaA / omega) * sin(omega * delta_t / 2.0f);
}


/************************************************************************************************************/
/** DQA BY DOMEGAA
* Ancillary function to calculate part of Jacobian which is repeatable due
* to symmetry. Here omegaA is one of omegax, omegay, omegaz and similarly with qA.
*/
double CMonoSlamMotionModel::dqA_by_domegaA(double omegaA, double omega)
{
  return (delta_t / 2.0f) * omegaA * omegaA / (omega * omega) * cos(omega * delta_t / 2.0f)
    + (1.0f / omega) * (1.0f - omegaA * omegaA / (omega * omega)) * sin(omega * delta_t / 2.0f);
}


/************************************************************************************************************/
/** DQA BY DOMEGAB
* Ancillary function to calculate part of Jacobian which is repeatable due to symmetry.
* Here omegaB is one of omegax, omegay, omegaz and similarly with qA.
*/
double CMonoSlamMotionModel::dqA_by_domegaB(double omegaA, double omegaB, double omega)
{
  return (omegaA * omegaB / (omega * omega)) *
         ( (delta_t / 2.0f) * cos(omega * delta_t / 2.0f)
         - (1.0f / omega) * sin(omega * delta_t / 2.0f) );
}

/************************************************************************************************************/
  void CMonoSlamMotionModel::dq3_dq1(mrpt::math::CQuaternion<double>& q2,CMatrixDouble &M)	//matrix(4x4)
{
  M.setSize(4,4);
  M(0,0)=q2.r();	M(0,1)=-q2.x();		M(0,2)=-q2.y();		M(0,3)=-q2.z();
  M(1,0)=q2.x();	M(1,1)=q2.r();		M(1,2)=-q2.z();		M(1,3)=q2.y();
  M(2,0)=q2.y();	M(2,1)=q2.z();		M(2,2)=q2.r();		M(2,3)=-q2.x();
  M(3,0)=q2.z();	M(3,1)=-q2.y();		M(3,2)=q2.x();		M(3,3)=q2.r();
}

/************************************************************************************************************/
void CMonoSlamMotionModel::dq3_dq2(mrpt::math::CQuaternion<double>& q1,CMatrixDouble &M)	//matrix(4x4)
{
  M.setSize(4,4);
  M(0,0)=q1.r();	M(0,1)=-q1.x();		M(0,2)=-q1.y();		M(0,3)=-q1.z();
  M(1,0)=q1.x();	M(1,1)=q1.r();		M(1,2)=q1.z();		M(1,3)=-q1.y();
  M(2,0)=q1.y();	M(2,1)=-q1.z();		M(2,2)=q1.r();		M(2,3)=q1.x();
  M(3,0)=q1.z();	M(3,1)=q1.y();		M(3,2)=-q1.x();		M(3,3)=q1.r();
}
/************************************************************************************************************/
