/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Perception and Robotics               |
   |      research group, University of Malaga (Spain).                        |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/core.h>
#include <mrpt/monoslam/CMonoSlam.h>
#include <mrpt/monoslam/CMonoSlamMotionModel.h>

using namespace std;
using namespace mrpt;
using namespace mrpt::utils;
using namespace mrpt::poses;
using namespace mrpt::slam;
using namespace mrpt::monoslam;
using namespace mrpt::math;
using namespace mrpt::gui;
using namespace mrpt::vision;
using namespace mrpt::hwdrivers;


/******************************************************************************************************************************/
/*************************************************     EKF FUNCTIONS      *****************************************************/

void  CMonoSlam::function_fv(const CMatrixDouble &u, mrpt::math::CVectorDouble &xv)
{
	//In MonoSlam we use the action 'u' as delta_t (time between frames)
	ASSERT_(size(u,1)==1 && size(u,2)==1)  // u must be 1x1 -> the time between two frames
	mmm.set_delta_t(u(0,0));
	mmm.func_fv(xv);
}

void  CMonoSlam::function_F(CMatrixDouble &F_parcial)
{
	mrpt::math::CVectorDouble xv;
	xv.resize(state_size);
	m_xkk.extract_vector(0,xv);

		
	mmm.func_dfv_by_dxv(xv,F_parcial);
}

void  CMonoSlam::function_Q(CMatrixDouble &Q)
{
	mrpt::math::CVectorDouble xv;
	xv.resize(state_size);
	m_xkk.extract_vector(0,xv);

	mmm.func_Q(xv,Q);
}

void  CMonoSlam::function_h(CMatrixDouble &h, CMatrixDouble &Hx_Hy)
{
 	size_t nLMs = (m_xkk.size() - get_vehicle_size() ) / get_feature_size();

    size_t  vehicle_size = get_vehicle_size();
    size_t  obs_size  = get_observation_size();
    size_t  feature_size = get_feature_size();

	h_calculate(h,Hx_Hy);
}

void  CMonoSlam::function_R(CMatrixDouble &h, CMatrixDouble &R)
{
	// R is the noise variance matrix(pag3 A Visual Compass based on SLAM. Zaragoza)
 
	size_t	nObs = h.getRowCount();
	if (nObs>0)
	{
 		size_t  O = get_observation_size();  // size of each observation

		R.setSize(nObs*O, O );
		R.zeros();

		for (size_t obs=0;obs<nObs;obs++)
		{
			if(h(obs,0)>0)
			{
				size_t  row_off = O*obs;
				for (size_t i=0;i<O;i++)
					for (size_t j=0;j<O;j++)
						R.set_unsafe( row_off+i,j, i==j ? square(options.sigma_image_noise) : 0 );
			}
		}
	}
	else
	{

		THROW_EXCEPTION("[MRPT-MonoSlam-function_R] landmarks # is NULL!");

	}

}


void  CMonoSlam::normalize()
{
		//Normalize the quaternion
		CQuaternionDouble q;
		m_xkk.extract_vector(3,q);
		q.q_normalise();
		m_xkk.insert_vector(3,q);

		//normalize covariance matrix (Pxx)
		CMatrixDouble Jnorm(state_size,state_size);
		CMatrixDouble temp(4,4);
		Jnorm.unit();
/********************************************  MONTIEL Method ************************************/
		q.q_normJac(temp);
/*******************************************  DAVISON Method ************************************/
//		temp = dqnorm_by_dq(q);
/********************************************************************************************/
		Jnorm.insertMatrix(3,3,temp);
		CMatrixDouble aux, aux2;
		m_pkk.extractMatrix(0,0,state_size,state_size, aux);

		Jnorm.multiply_HCHt(aux, aux2);
		m_pkk.insertMatrix(0,0,aux2); //Jnorm * aux * ~Jnorm);

		//normalize covariance matrix (Pxy)
		if (m_pkk.getRowCount()>state_size)
			for (unsigned int i = 0 ; i < (m_xkk.size()-state_size ) / feature_size ; i++)
			{
			    m_pkk.extractMatrix(0,state_size+i*feature_size,state_size,feature_size,aux);

				aux2.multiply(Jnorm, aux);
				m_pkk.insertMatrix(0,state_size+i*feature_size,aux2);

				//(~pkk).extractMatrix(0,state_size+i*feature_size,state_size,feature_size,aux);
				//m_pkk.insertMatrix(state_size+i*feature_size,0,aux);
				m_pkk.insertMatrix(state_size+i*feature_size,0,~aux2);

			}

}

/*************************************************   Functions Used in UKF   *****************************************************/
void  CMonoSlam::h_UKF(CMatrixDouble &h)
{
	// h contiene la posicion (en pixeles sin distorsionar) que tienen todas las caracteristicas que se ven en la imagen
	// H es el matriz que contiene las 2 ecualiones de medidas linelizadas para cada caracteristica

	//Cada caracterstica tiene 6 parametros (x,y,z) de la posicion de la camara y (hx,hy,hz) es un vector que describe su direccion

	unsigned int numFeatures = (unsigned int)(m_xkk.size() - state_size) / feature_size;

	mrpt::math::CVectorDouble XfeatToPredict;	XfeatToPredict.resize(feature_size);
	CMatrixDouble zi(1,2);

    //Inicializacion h_all
	h.setSize(numFeatures,2);
	h.ones();
	h *=-1;

    for (unsigned int i = 0; i<numFeatures ; i++)
	{
        m_xkk.extract_vector(state_size+i*feature_size,XfeatToPredict);

        // Measurement equation
		// Devuelve la posicion del pixel distorsionado sobre la imagen en la que deberia estar la caracteristica buscada
		hi( XfeatToPredict, zi,i );

		//In UKF we need all values, although any sample predict the feature out of vision range
		h.insertMatrix(i,0,zi);
	}
}

void  CMonoSlam::visibility(CMatrixDouble &h)
{
	unsigned int excluded_band = 11;
	for (size_t i = 0 ; i < h.getRowCount() ; i++)
		if (( h(i,0)<excluded_band ) || ( h(i,0)>cam.cam_ncols()-excluded_band ) || ( h(i,1)<excluded_band ) || ( h(i,1)>cam.cam_nrows()-excluded_band ))
		{
			h(i,0) = -1;
			h(i,1) = -1;
		}
}



//---------------------------------------------------------------------------
//------------------------ EKF Virtual methods  -----------------------------
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
/** Must return the action vector u.
  * \param out_u The action vector which will be passed to OnTransitionModel
  */
//---------------------------------------------------------------------------

void CMonoSlam::OnGetAction( CVectorTemplate<KFTYPE> &out_u )
{
	MRPT_TRY_START

	// The "action" is just the time step between two frames:
	out_u.resize(1);
	out_u[0] = mmm.get_delta_t();
//
//	if (sim_or_rob!=0)		// real-time processing
//	{
//		out_u.resize(1);
//		out_u[0] = mmm.get_delta_t();//m_dt;
//	}
//	else
//	{     //Using rawlogs
// 
//#if USE_ONLY_RAWLOG_IMAGE
//
//		out_u.resize(1);
//		out_u[0] = mmm.get_delta_t();//m_dt;
//#else
//
//		out_u.resize(6);
// 
//		// Get odometry estimation:
//		CActionRobotMovement2DPtr actMov2D = m_action->getBestMovementEstimation();
//		CActionRobotMovement3DPtr actMov3D = m_action->getActionByClass<CActionRobotMovement3D>();
//		if (actMov3D)
//		{
//			actMov3D->poseChange.mean.getAsVector(out_u);
//		}
//		else
//		if (actMov2D)
//		{
//			CPose2D estMovMean;
//			actMov2D->poseChange->getMean(estMovMean);
//
//			out_u[0]= estMovMean.x();
//			out_u[1]= 0;
//			out_u[2]= estMovMean.y();
//
//			//MAAA: Test
//			out_u[3]= estMovMean.phi(); //yaw
//			out_u[4]= 0;
//			out_u[5]= 0;
//		    //std::cout << " -------------fYawN  deg  :"<< RAD2DEG(estMovMean.phi()) << "\n"; 
//		}
//		else
//		{
//			// Left u as zeros
//			for (size_t i=0;i<6;i++) out_u[i]=0;
//		}
//
//#endif 
//
//	}//else
//
	MRPT_TRY_END
}


//---------------------------------------------------------------------------
/** Implements the transition model \f$ \hat{x}_{k|k-1} = f( \hat{x}_{k-1|k-1}, u_k ) \f$
  * \param in_u The vector returned by OnGetAction.
  * \param inout_x At input has \f[ \hat{x}_{k-1|k-1} \f] , at output must have \f$ \hat{x}_{k|k-1} \f$ .
  * \param out_skip Set this to true if for some reason you want to skip the prediction step (to do not modify either the vector or the covariance). Default:false
  */
//---------------------------------------------------------------------------
void CMonoSlam::OnTransitionModel(
	const CVectorTemplate<KFTYPE>& in_u,
	CVectorTemplate<KFTYPE>&       inout_x,
	bool &out_skipPrediction
	 )
{
	MRPT_TRY_START
	
	// Do not update the vehicle pose & its covariance until we have some landmakrs in the map,
	// otherwise, we are imposing a lower bound to the best uncertainty from now on:
	if (m_xkk.size() == get_vehicle_size() )
	{
		out_skipPrediction = true;
		return;
	}

	ASSERT_(inout_x.size()>=6);

    //std::cout << " inout_x  at the begining of the transition method applyment:"<< inout_x << "\n"; 

	ASSERT_(in_u.size()==1 || in_u.size()==6);

	if (in_u.size()==1) // We dont have any kind of odometry:
	{
		ASSERT_(in_u.size()==1); // u[0] is the time between frames

		CMatrixDouble u(1,1);
		u(0,0) = in_u[0];
		function_fv(u,inout_x);

		CQuaternionDouble q;
		m_xkk.extract_vector(3,q);
		CMatrixDouble Rrw(3,3),Rwr(3,3);
		q.q_rotation_matrix(Rwr);
		double fYaw,fPitch,fRoll;//RAD
		q.rpy(fRoll, fPitch, fYaw);	

 		CPose3D  robotPosekk(m_xkk[0],m_xkk[1],m_xkk[2],RAD2DEG(fYaw),RAD2DEG(fPitch),RAD2DEG(fRoll));
		m_robotPose = robotPosekk;

	}
	else   // We have odometry of some kind
	{
		ASSERT_(in_u.size()==6); // 6D odometry + dt

		//MAAA: get euler angles from quat.
		CQuaternionDouble q; 
		inout_x.extract_vector(3,q);
		CMatrixDouble Rrw(3,3),Rwr(3,3);
		q.q_rotation_matrix(Rwr);
		Rrw = Rwr.inv();

 		//std::cout << " qk                   :"<< q  << "\n"; 

		double fYaw,fPitch,fRoll;//RAD
		q.rpy(fRoll, fPitch, fYaw);	
		
 		CPose3D  robotPosek(inout_x[0],inout_x[1],inout_x[2],fYaw,fPitch,fRoll);

		//MAAA: order of chanels need testing
		CPose3D  odoIncrement(in_u[0],in_u[1],in_u[2],in_u[3] /*yaw*/,in_u[4],in_u[5]);

 		// Pose composition:
		CPose3D  robotPose = robotPosek + odoIncrement;
		m_robotPose = robotPose;
		 
		inout_x[0]=m_robotPose.x();
		inout_x[1]=m_robotPose.y();
		inout_x[2]=m_robotPose.z();

		CQuaternionDouble qNew;
  		robotPose.getAsQuaternion(qNew);
		inout_x[3]= (qNew)[0];
		inout_x[4]= (qNew)[1];
		inout_x[5]= (qNew)[2];
		inout_x[6]= (qNew)[3];
  	    qNew.q_rotation_matrix(Rwr);
		////Rrw = Rwr.inv();

		//std::cout << " qNew from robotPose  :"<< qNew << "\n"; 

		// Get odometry estimation:
		CActionRobotMovement2DPtr actMov2D = m_action->getBestMovementEstimation();
		CActionRobotMovement3DPtr actMov3D = m_action->getActionByClass<CActionRobotMovement3D>();
		if (actMov3D)
		{

			//MAAA:TODO
 		}
		else if (actMov2D)
		{				
			if (actMov2D->hasVelocities)
			{
				//std::cout << format(" Velocity info: v=%.03f m/s  w=%.03f deg/s\n", actMov2D->velocityLin, RAD2DEG(actMov2D->velocityAng) );
				mrpt::math::CVectorDouble vLocal(3),vW(3);
                vLocal[0] = actMov2D->velocityLin;
		        vLocal[1] = 0;
	            vLocal[2] = 0;

 				//vW = Rwr * vLocal;
				Rwr.multiply_Ab(vLocal, vW); 

				inout_x[7]= vW[0];
				inout_x[8]= vW[1];
				inout_x[9]= vW[2];

				inout_x[10]= 0;
				inout_x[11]= (actMov2D->velocityAng);
				inout_x[12]=  0;		
			}
	 

 		}

		//MAAA:Test, OK
		//CQuaternionDouble qN(inout_x[3],inout_x[4],inout_x[5],inout_x[6]);
  //	    qN.rpy(fRoll, fPitch, fYaw);	
		//CPose3D  robotPoseT(inout_x[0],inout_x[1],inout_x[2],fYaw,fPitch,fRoll);
		std::cout << " Actions   :"<< in_u << "\n"; 
		//std::cout << " XvNew     :"<< inout_x << "\n"; 
		//std::cout << " RobotPose0:"<< robotPose0 << "\n"; 
		//std::cout << " RobotPoseT:"<< robotPoseT << "\n"; 
		std::cout << " RobotPoseN:"<< robotPose << "\n"; 
		std::cout << " m_robPose :"<< m_robotPose << "\n"; 
		//std::cout << " -------------fYawN  deg  :"<< RAD2DEG(Yaw_New) << "\n"; 
		//std::cout << " -------------fRollN  deg :"<< RAD2DEG(Roll_New) << "\n"; 

 		std::cout << " inout_x  at the end of Transition applyment:"<< inout_x << "\n"; 

 	}


	MRPT_TRY_END

}

//---------------------------------------------------------------------------
/** Implements the transition Jacobian \f$ \frac{\partial f}{\partial x} \f$
  * \param out_F Must return the Jacobian.
  *  The returned matrix must be \f$N \times N\f$ with N being either the size of the whole state vector or get_vehicle_size().
  */
//---------------------------------------------------------------------------
void CMonoSlam::OnTransitionJacobian(
	CMatrixTemplateNumeric<KFTYPE> & out_F
	 )
{
	MRPT_TRY_START

	function_F(out_F);
	out_F.setSize(state_size,state_size);

//	{
//#if USE_ONLY_RAWLOG_IMAGE	
//
//		function_F(out_F);
//		out_F.setSize(state_size,state_size);
//	
//#else
//		
//		double delta_t = mmm.get_delta_t();
//		out_F.setSize(state_size,state_size);
//        out_F.unit();
//
//		// Fill in F = I * delta_t
//		CMatrixDouble Temp33A(3,3);
//		for (int i=0;i<3;i++)
//			Temp33A(i,i)=delta_t;
//
//		out_F.insertMatrix(0,7,Temp33A);	// (0,7) is the place for the "I * delta_t" 3x3 submatrix
//	    
//		// Fill in dqnew_by_dq
//		// Warning: This function must use x(k|k), not x(k+1|k) --> Beware of not calling it after "func_fv", which modified "x_k_k"
//		CVectorDouble w_old(3); 
//		CQuaternionDouble q_old;
//		m_xkk.extract_vector(10,w_old);
//		m_xkk.extract_vector(3,q_old);
//		CQuaternionDouble qwt(w_old*delta_t);
//
//		CMatrixDouble Temp44A(4,4);
//		Temp44A.zeros();
//		mmm.dq3_dq2(qwt,Temp44A);
//
//		out_F.insertMatrix(3,3,Temp44A);
//
//		//// Fill in dqnew_by_domega = d(q x qwt)_by_dqwt . dqwt_by_domega
//		mmm.dq3_dq1(q_old,Temp44A); // Temp44A is d(q x qwt) by dqwt
//
//		// Use function below for dqwt_by_domega
//		CMatrixDouble Temp43A(4,3);
//		Temp43A.zeros();
//		mmm.dqomegadt_by_domega(w_old, Temp43A);
//
//		// Multiply them together
//		CMatrixDouble Temp43B(4,3);
//		Temp43B	= Temp44A * Temp43A;
//		// And plug it in
//		out_F.insertMatrix(3,10,Temp43B);
//#endif
//
//
//
// 	}
//

	//if(m_verbose)
	//  std::cout << "TransitionJacobian F: \n" << out_F;
	//if (options.save_data_files)
	//	out_F.saveToTextFile("./monoSlamOut/df_dxv.txt");		

	MRPT_TRY_END
}


//---------------------------------------------------------------------------
/** Implements the transition noise covariance \f$ Q_k \f$
  * \param out_Q Must return the covariance matrix.
  *  The returned matrix must be of the same size than the jacobian from OnTransitionJacobian
  */
//---------------------------------------------------------------------------
void CMonoSlam::OnTransitionNoise(
	CMatrixTemplateNumeric<KFTYPE> & out_Q
	 )
{
	MRPT_TRY_START
	function_Q(out_Q);
 	
	//std::cout << "[Mono-Slam]  m_pkk:"<< m_pkk << "\n"; 
	MRPT_TRY_END
}


//---------------------------------------------------------------------------
/** This is called between the KF prediction step and the update step to allow the application to employ the prior distribution of the system state in the detection of observations (data association), where applicable.
  * \param out_z A \f$N \times O\f$ matrix, with O being get_observation_size() and N the number of "observations": how many observed landmarks for a map, or just one if not applicable.
  * \param out_R A matrix of size NOxO (O being get_observation_size() and N the number of observations). It is the covariance matrix of the sensor noise for each of the returned observations. The order must correspond to that in out_z.
  * \param out_data_association An empty vector or, where applicable, a vector where the i'th element corresponds to the position of the observation in the i'th row of out_z within the system state vector, or -1 if it is a new map element and we want to insert it at the end of this KF iteration.
  *  This method will be called just once for each complete KF iteration.
  *  \note Since MRPT 0.5.5, the method "OnGetObservations" returning R as a matrix is called instead from the Kalman Filter engine. However, for compatibility, a default virtual method calls this method and build the R matrix from the diagonal R2.
  */
//---------------------------------------------------------------------------
void CMonoSlam::OnGetObservations(
	CMatrixTemplateNumeric<KFTYPE> &out_z,
	CMatrixTemplateNumeric<KFTYPE> &out_R,
	vector_int                     &out_data_association
	)
{
	MRPT_TRY_START

	// out_z: NxO  (N:#obs, O:2)
	// data_association: N

	/** JL says:
		OnGetObservations: This is the ideal place for generating observations in those 
	    applications where it requires an estimate of the current state (e.g. in MonoSLAM, 
		to predict where each landmark will be found onto the image). At this point the internal 
		state vector and covariance contain the "prior distribution", i.e. updated through the 
		transition model. This is also the place where data association must be solved (in mapping problems). 
	*/

	 // Calulate jacobians
     function_h( m_h, m_Hx_Hy);

	 out_z.setSize(m_h.getRowCount(),m_h.getColCount());
	 out_z = m_h;

	 // Prepare observed feature list 
	 m_ObservedfeatList.clear();

	 //out_z.setSize(0,0);
	 for(size_t i=0;i<m_h.getRowCount();i++)
	 {
		 if(m_h(i,0)>=0 && m_h(i,1)>=0)
		 {
	    //   out_z.setSize(out_z.getRowCount()+1,m_h.getColCount());
		   //out_z(out_z.getRowCount()-1,0) = m_h(i,0);
		   //out_z(out_z.getRowCount()-1,1) = m_h(i,1);

		   //debug mode
#if defined(_DEBUG) || (MRPT_ALWAYS_CHECKS_DEBUG)
		   ASSERT_(m_featList.size()>i)
#endif
			   m_ObservedfeatList.insert(m_featList[i]);
                //THROW_EXCEPTION("[MRPT-MonoSlam] Feature list size is greater than the landmarks'");
		 }
 	 }

	// Set data association: each index in the currently observed features has a 
	//  corresponding index in map's features:
	//const size_t N = out_z.getRowCount();
  	const size_t num_feats_in_map = (m_xkk.size() - state_size) / feature_size;

 	//Get R :
    function_R( out_z/*m_h*/, out_R);

	//Get S
	//MAAA: TODO; add a getter to the EKF base class to get this "m_S" from there!
	if(m_iterations_count<2)
	{
		// S = H P ~H + R
		m_S.setSize(0,0);
		m_Hx_Hy.multiply_HCHt(m_pkk,m_S);

        //std::cout<<"[Mono-SLAM] m_S: "<<m_S<<"\n";
		const size_t obs_size  = get_observation_size();
		size_t indx = 0;
		for (size_t i=0;i<num_feats_in_map;i++)
		{					
			if(m_h(i,0)>0 && m_h(i,1)>0)
			{
				size_t  obs_idx_off = i*obs_size;
				for (size_t j=0;j<obs_size;j++)
					for (size_t k=0;k<obs_size;k++)
						m_S(obs_idx_off+j,obs_idx_off+k) += out_R(j,k);
				//indx++;
			}
		}
	}
 
 	//--------------------------------------------------------------------
	// Find Matches
    //--------------------------------------------------------------------


#if 1
	
	//reloj.Tic();
	if (options.verbose)
		std::cout<<"matching\n";

		//std::cout<<"[Mono-Slam] m_h : \n"<<m_h<<"\n";
		//std::cout<<"[Mono-Slam] m_S : \n"<<m_S<<"\n";
		//std::cout<<"[Mono-Slam] m_Hx_Hy : \n"<<m_Hx_Hy<<"\n";

	if ( KF_options.method == mrpt::bayes::kfEKFNaive || KF_options.method == mrpt::bayes::kfIKFFull )
		CMonoSlam::MSMatching(m_h,m_S,m_Hx_Hy,m_frame,
			                  m_hh /*out matched/predicted uv by projection*/,
							  m_zz /*out matched/matched uv */,
							  m_HH,m_RR,m_FeatPatch);
	else
		CMonoSlam::MSMatching(m_h,m_S,m_frame,m_hh,m_zz,m_RR,m_match,m_FeatPatch);

		//std::cout<<"[Mono-Slam] m_HH : \n"<<m_HH<<"\n";
		//std::cout<<"[Mono-Slam] m_RR : \n"<<m_RR<<"\n";

    //std::cout<<"matching: "<<reloj.Tac()<<"\n";

#endif


	const size_t nMatchedData = m_zz.getRowCount()/2;
	out_data_association.clear();
 	out_data_association.resize(nMatchedData,-1);

	int index=0;
	int indexZ=0;
    for (unsigned int i = 0; i< num_feats_in_map; i++)
	{
  		if (m_MatchCoord(i,0) > -1)//if is matched
		{
 		  out_data_association[index] = i;
		  index++;
		}
	}

 	std::cout<<"out_data_association: "<<out_data_association<<"\n";
 	std::cout<<"m_MatchCoord: \n"<<m_MatchCoord<<"\n";

	//-------------------------------------------------------------------------
 	// ELIMINATE Ambiguity
	//-------------------------------------------------------------------------

#if 0	
	//reloj.Tic();
	if (m_type == 0)
		CMonoSlam::Ambiguity(m_HH,m_RR,m_hh,m_zz);
	else
		CMonoSlam::Ambiguity(m_RR,m_hh,m_zz,m_match);

	//std::cout<<"hh-zz: "<<~hh-~zz<<"\n";
	//std::cout<<"ambiguas: "<<reloj.Tac()<<"\n";
#endif


	MRPT_TRY_END

}



/** Implements the observation prediction \f$ h_i(x) \f$ and the Jacobians \f$ \frac{\partial h_i}{\partial x} \f$ and (when applicable) \f$ \frac{\partial h_i}{\partial y_i} \f$.
  * \param in_z This is the same matrix returned by OnGetObservations(), passed here for reference.
  * \param in_data_association The vector returned by OnGetObservations(), passed here for reference.
  * \param in_full If set to true, all the Jacobians and predictions must be computed at once. Otherwise, just those for the observation in_obsIdx.
  * \param in_obsIdx If in_full=false, the row of the observation (in in_z and in_data_association) whose innovation & Jacobians are to be returned now.
  * \param out_innov The difference between the expected observation and the real one: \f$ \tilde{y} = z - h(x) \f$. It must be a vector of length O*N (in_full=true) or O (in_full=false), with O=get_observation_size() and N=number of rows in in_z.
  * \param out_Hx One or a vertical stack of \f$ \frac{\partial h_i}{\partial x} \f$.
  * \param out_Hy An empty matrix, or one or a vertical stack of \f$ \frac{\partial h_i}{\partial y_i} \f$.
  *
  *  out_Hx must consist of 1 (in_full=false) or N (in_full=true) vertically stacked \f$O \times V\f$ matrices, and out_Hy must consist of 1 or N \f$O \times F\f$ matrices, with:
  *  - N: number of rows in in_z (number of observed features, or 1 in general).
  *  - O: get_observation_size()
  *  - V: get_vehicle_size()
  *  - F: get_feature_size()
  */
void CMonoSlam::OnObservationModelAndJacobians(
	const CMatrixTemplateNumeric<KFTYPE> &in_z,
	const vector_int                     &data_association,
	const bool                           &in_full,
	const int                            &in_obsIdx,
	CVectorTemplate<KFTYPE>       		 &ytilde,
	CMatrixTemplateNumeric<KFTYPE>       &Hx,
	CMatrixTemplateNumeric<KFTYPE>       &Hy )
{
	MRPT_TRY_START

	/* -------------------------------------------
		Hx : dh_dxv   -> Jacobian of the observation model wrt the robot pose
		Hy : dh_dyi   -> Jacobian of the observation model wrt each landmark mean position
		L=# of landmarks in the map (ALL OF THEM)
	  ------------------------------------------- */

 	size_t  obs_size  = get_observation_size();
 	size_t i,    nLMs = (m_xkk.size() - state_size) / feature_size;
 	size_t nVisibleLMs = in_z.getRowCount(); 

	// Mean of the prior of the robot pose:
	//CPose3D  robotPose( m_xkk[0],m_xkk[1],m_xkk[2],m_Yaw_New,m_Pitch_New,m_Roll_New );
	//CPose3D  sensorPoseOnRobot( 0,0,0,0,0,0 );

	// Requested to build just one LM?
	if (!in_full)
		nLMs = 1;

	ytilde.resize( obs_size*nLMs,0); //initial value = 0;

	Hx.setSize    ( obs_size*nLMs, state_size );
    Hy.setSize    ( obs_size*nLMs, feature_size );
 
	//They were calculated using CMonoSlam::OnGetObservations method
    //std::cout<<"[Mono-Slam] start Hx OK \n";

	m_Hx_Hy.extractMatrix(0,0,Hx);

    //std::cout<<"[Mono-Slam] Hx OK \n";

	if (in_full)
	   m_Hx_Hy.extractMatrix(0,state_size,Hy);
 	else
	   m_Hx_Hy.extractMatrix(in_obsIdx*obs_size,state_size,Hy);

    //std::cout<<"[Mono-Slam]   Hy  OK\n";

 	//CTicTac   tictac;
    int obsIdx = in_obsIdx;

	for (i=0;i<nLMs;i++)  // nLMs = 1 if full=false!
	{
		size_t idxOfFeatureInMap;
		if (in_full)
			idxOfFeatureInMap = i;
		else
		{
			ASSERT_(data_association[in_obsIdx]>=0)  // It is an already mapped feature!
			idxOfFeatureInMap = data_association[in_obsIdx];  // Just take one LM
		}
 
		//size_t  row_in = 2*idxOfFeatureInMap;
		size_t  row = 2*i;

		// ---------------------------------------------------
		// Generate the ytilde:
		// ---------------------------------------------------

		// Fill out "ytilde": If this LM has been observed, the usual value. Otherwise, zeroes:
		int idxThisLMInZ = -1;
		for (size_t p=0;p< data_association.size();p++)
		{
			if ( data_association[p] == static_cast<int>(idxOfFeatureInMap))
			{
				idxThisLMInZ = p;
				break;
			}
		}

		if (idxThisLMInZ<0)
		{
			// Not observed:
			ytilde[row+0] =
			ytilde[row+1] = 0;
		}
		else //if (!(m_zz.getRowCount()==0 || m_hh.getRowCount()==0))
		{
	    	// Yes: compute ytilde: in_z == m_h
  		//		ytilde[row+0] = (m_zz(idxThisLMInZ,0) - m_hh(idxThisLMInZ,0));//u
				//ytilde[row+1] = (m_zz(idxThisLMInZ+1,0) - m_hh(idxThisLMInZ+1,0));//v
  				ytilde[row+0] = ( m_z(idxOfFeatureInMap,0) - m_h(idxOfFeatureInMap,0));//u
				ytilde[row+1] = ( m_z(idxOfFeatureInMap,1) - m_h(idxOfFeatureInMap,1));//v
              
				std::cout<<"[Mono-Slam] idxOfFeatureInMap  :"<< idxOfFeatureInMap << "\n";
 		}

 
	}
 
	//if(m_verbose)
	//{
       std::cout<<"[Mono-Slam] ytilde: \n "<< ytilde << "\n";
       //std::cout<<"[Mono-Slam] zz: \n"<< m_zz << "\n";
       //std::cout<<"[Mono-Slam] hh: \n"<< m_hh << "\n";
       //std::cout<<"[Mono-Slam] Hx: \n"<< Hx << "\n";
      // std::cout<<"[Mono-Slam] Hy: \n"<< Hy << "\n";
       //std::cout<<"[Mono-Slam] m_Hx_Hy:row,col " << m_Hx_Hy.getRowCount() << ", " << m_Hx_Hy.getColCount() << "\n";
	   //std::cout<<"[Mono-Slam] Hx: row,col " << Hx.getRowCount() << ", " << Hx.getColCount() << "\n";
	//}

	MRPT_TRY_END

}



/** If applicable to the given problem, this method implements the inverse observation model needed to extend the "map" with a new "element".
  * \param in_z This is the same matrix returned by OnGetObservations().
  * \param in_obsIndex The index of the observation whose inverse sensor is to be computed. It corresponds to the row in in_z where the observation can be found.
  * \param in_idxNewFeat The index that this new feature will have in the state vector (0:just after the vehicle state, 1: after that,...). Save this number so data association can be done according to these indices.
  * \param out_yn The F-length vector with the inverse observation model \f$ y_n=y(x,z_n) \f$.
  * \param out_dyn_dxv The \f$F \times V\f$ Jacobian of the inv. sensor model wrt the robot pose \f$ \frac{\partial y_n}{\partial x_v} \f$.
  * \param out_dyn_dhn The \f$F \times O\f$ Jacobian of the inv. sensor model wrt the observation vector \f$ \frac{\partial y_n}{\partial h_n} \f$.
  *
  *  - O: get_observation_size()
  *  - V: get_vehicle_size()
  *  - F: get_feature_size()
  */

void  CMonoSlam::OnInverseObservationModel(
	const CMatrixTemplateNumeric<KFTYPE> &in_z,
	const size_t                         &in_obsIdx,
	const size_t                         &in_idxNewFeat,
	CVectorTemplate<KFTYPE>              &out_yn,
	CMatrixTemplateNumeric<KFTYPE>       &out_dyn_dxv,
	CMatrixTemplateNumeric<KFTYPE>       &out_dyn_dhn )
{
	MRPT_TRY_START

	//this task has been done inside the go_one_step method
	//adding new feature
	//THROW_EXCEPTION("Inverse sensor model required but not implemented in derived class.")


    MRPT_TRY_END

}


 void  CMonoSlam::OnPostIteration()
 {

	MRPT_TRY_START

	if (options.verbose)
	{
		std::cout<<"SLAM state after 'update'" << endl;
		CQuaternionDouble q1;
		m_xkk.extract_vector(3,q1);
		double r,p,y;
		q1.rpy(r,p,y);
		std::cout<<"X: "<<m_xkk[0]<<"		    Y: "<<m_xkk[1]<<"		Z: "<<m_xkk[2]<<"\n";
		std::cout<<"R: "<<(180.0/M_PI)*r<<"		P: "<<(180.0/M_PI)*p<<"		Y: "<<(180.0/M_PI)*y<<"\n";
		std::cout<<"vel x: "<<m_xkk[7]<<"	vel y: "<<m_xkk[8]<<"	vel z: "<<m_xkk[9]<<"\n";
		std::cout<<"velw x: "<<m_xkk[10]<<"	velw y: "<<m_xkk[11]<<"	velw z: "<<m_xkk[12]<<"\n\n";
 	}

	if (options.save_data_files)
	{
		CImage	patch;
		for (unsigned int tpm=0;tpm<m_FeatPatch.size();tpm++)
		{
			std::cout << "index: " << tpm <<"	nombre: "<< m_FeatPatch.get_name(tpm) <<"	hits: "<<m_FeatPatch.get_hits(tpm)<<"\n";
			m_FeatPatch.get_patch(tpm,patch);
			patch.saveToFile( format("./monoSlamOut/p%01d.bmp",tpm) );
		}

		m_pkk.saveToTextFile("./monoSlamOut/pkkupdatec.txt");
		m_xkk.saveToTextFile("./monoSlamOut/xkkupdatec.txt");
 
		m_hh.saveToTextFile("./monoSlamOut/m_hh.txt");
		m_zz.saveToTextFile("./monoSlamOut/m_zz.txt");
 
		m_h.saveToTextFile("./monoSlamOut/m_h.txt");
		m_S.saveToTextFile("./monoSlamOut/m_S.txt");
	    //if (m_type == 0)
		{
			m_HH.saveToTextFile("./monoSlamOut/m_HH.txt");
			m_Hx_Hy.saveToTextFile("./monoSlamOut/Hx_Hy.txt");
		}
	}
 
 
	//[x y z q1 q2 q3 q4  vx vy vz wx wy wz fYaw,fPitch,fRoll,time ]
	// 0 1 2 3  4  5  6  7  8  9  10 11  
	m_simTime += mmm.get_delta_t();
 	CVectorDouble xv,orient(3);
	orient[0] = m_robotPose.yaw();orient[1] = m_robotPose.pitch();orient[2] = m_robotPose.roll();
	xv.resize(state_size);
 	m_xkk.extract_vector(0,xv);
	xv.resize(state_size+4);
	xv[xv.size()-1] = m_simTime;
	xv.insert_vector(state_size,orient);
	m_robotPath.setSize(m_robotPath.getRowCount()+1,state_size+4);
    m_robotPath.insertRow(m_robotPath.getRowCount()-1,xv);


	//   CMatrixDouble pKKnew;
	//m_pkk.extractMatrix( 0/*size(m_pkk,1)-3*/,size(m_pkk,2)-6,6,6, pKKnew);
	//std::cout << "[Mono-Slam]  pKKnew: \n"<< pKKnew  << "\n"; 
	//cin.get();

	MRPT_TRY_END
}


