/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Perception and Robotics               |
   |      research group, University of Malaga (Spain).                        |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/core.h>
#include <mrpt/monoslam/CMonoSlam.h>
#include <mrpt/monoslam/CMonoSlamMotionModel.h>

using namespace std;
using namespace mrpt;
using namespace mrpt::utils;
using namespace mrpt::poses;
using namespace mrpt::slam;
using namespace mrpt::monoslam;
using namespace mrpt::math;
using namespace mrpt::gui;
using namespace mrpt::vision;
using namespace mrpt::hwdrivers;


const size_t CMonoSlam::state_size = CMonoSlamMotionModel::state_size;	//!< State Vector Size
const size_t CMonoSlam::feature_size = CMonoSlamMotionModel::feature_size; //!< Feature Vector Size


/* --------------------------------------------------------------------------------------
	   Constructor
	-------------------------------------------------------------------------------------- */
CMonoSlam::CMonoSlam () :
	m_iterations_count	( 0 ),
	m_num_obj_visible	( 0 ),
	m_FeatPatch(),
	m_timeLastFrame	(INVALID_TIMESTAMP)
{
	mrpt::system::createDirectory( "./monoSlamOut" );

	this->resetFilterState();
}


/* --------------------------------------------------------------------------------------
	   resetFilterState
	-------------------------------------------------------------------------------------- */
void CMonoSlam::resetFilterState()
{
	m_iterations_count	= 0;
	m_num_obj_visible	= 0;

	prev_speed = 0;
	prev_xv.resize(13);

	// ------------------------------------------
	// covariance matrix initialization:
	//  A zeroed covariance SHOULD be OK with the maths:
	// ------------------------------------------
	m_pkk.zeros(13,13);

	// ------------------------------------------
	// state matrix initialization
	// m_xkk[0:2]: XYZ
	// m_xkk[3:6]: Quaternion
	// m_xkk[7:12]: xyz & quaternion velocities
	// ------------------------------------------
	m_xkk.assign(13,0);
	m_xkk[3] = 1.0;


	CQuaternionDouble q;
	CMatrixDouble Rrw(3,3),Rwr(3,3);
	m_xkk.extract_vector(3,q);
	//q.q_rotation_matrix(Rwr);
	//Rrw = Rwr.inv();

	double fYaw,fPitch,fRoll;//RAD
	q.rpy(fRoll, fPitch, fYaw);

	CPose3D  robotPose0(m_xkk[0],m_xkk[1],m_xkk[2],fYaw,fPitch,fRoll);
	m_robotPose = robotPose0;


	//[x y z q1 q2 q3 q4 vx vy vz wx wy wz fYaw fPitch fRoll, time ]
	// 0 1 2 3  4  5  6  7  8  9  10 11 12
 	m_simTime = 0;
	CVectorDouble xv,orient(3);
	orient[0] = m_robotPose.yaw();orient[1] = m_robotPose.pitch();orient[2] = m_robotPose.roll();
	xv.resize(state_size);
 	m_xkk.extract_vector(0,xv);
	xv.resize(state_size+4);
	xv[xv.size()-1] = m_simTime;
	xv.insert_vector(state_size,orient);
	m_robotPath.setSize(1,state_size+4);
    m_robotPath.insertRow(0,xv);

	//m_MatchCoord.setSize(1,1);

	m_FeatPatch.clear();
	m_timeLastFrame	 = INVALID_TIMESTAMP;
}

/*----------------------------------------------------------------
			Dtor
  ---------------------------------------------------------------- */
CMonoSlam::~CMonoSlam()
{
}


//----------------------------------------------------------------
/* One time initialization of the CMonoSlamInterface */
//----------------------------------------------------------------

//void  CMonoSlam::initialize(const CSensoryFrame &sensoryframe,
// 									 unsigned int type,
//									 bool show2D, bool show3D,
//									 bool verbose, bool savedatas)
//{
////	UTILS::CTicTac	reloj;
//	//CObservation		*observ;
//
//	m_type      = type;
//	m_show2D    = show2D;
//	m_show3D    = show3D;
//	m_verbose   = verbose;
//	m_savedatas = savedatas;
//
//	if (show2D && !wind2D)
//	{
//		wind2D = CDisplayWindowPtr( new CDisplayWindow(std::string("2D View")) );
//		wind2D->setPos(600,0);
//	}
//	if (show3D && !wind3D)
//	{
//		wind3D = CDisplayWindow3DPtr( new CDisplayWindow3D("3D WORLD MONOSLAM"/*,500,500*/) );
//	}
//
//	if (wind3D){
//	   world = wind3D->get3DSceneAndLock();
//		wind3D->setPos(0,0);
//	}
//
//	if (sim_or_rob!=0)		//Use a real-time camera
//	{
// 		// Take a few images to assure the camera is initialized:
//		cout<< "[mono-slam] Discarding first 3 seconds of images from camera..." << endl;
//		CTicTac timInit;
//		timInit.Tic();
//		while (timInit.Tac()<3.0)
//		{
//			m_camera->doProcess();
//		}
//
//		// Recover grabbed images:
//		CGenericSensor::TListObservations  lstCamObs;
//
//		m_camera->getObservations( lstCamObs );
//		if (lstCamObs.empty())
//			THROW_EXCEPTION("Specified camera seems not to grab images.")
//
//		ASSERT_( IS_CLASS( lstCamObs.rbegin()->second, CObservationImage ) );
//
//		//Set dt=0 at first
//  		m_DeltaTimeCounter.Tic();
//
//		// Get first image:
//		CObservationImagePtr imgp = CObservationImagePtr(lstCamObs.rbegin()->second);
//		initial_frame = imgp->image;
//
//		std::cout<< "[mono-slam] Initial frame: " << initial_frame.getHeight()<<","<<initial_frame.getWidth()<< endl;
//
//	}
//	else
//	{	// Off-line: Use a rawlog file
//
// 		CObservationStereoImagesPtr sImgs;
//		CObservationImagePtr Img;
//
//		sImgs = sensoryframe.getObservationByClass<CObservationStereoImages>();
//
//		if (sImgs)
//		{
//			initial_frame = sImgs->imageLeft;
//		}
//		else
//		{
//			Img = sensoryframe.getObservationByClass<CObservationImage>();
//			ASSERT_(Img);
//			initial_frame = Img->image;
//		}
//
//		//CSensoryFrame::iterator o= sensoryframe.begin();
//		m_dFrameTimeStamp = sensoryframe.begin()->pointer()->timestamp;
//
// 	}
//
//	if (options.verbose)
//		std::cout<<"[mono-slam] Initializing features..." << endl;
//
//	initial_frame.saveToFile("./monoSlamOut/frame_initial.jpg");
//
//	////initial_frame.grayscaleInPlace();
//	if(m_bGrayScaleProcessing)
//	  initial_frame = initial_frame.grayscale();
//
//	if(m_bScaleHalf)
//		initial_frame = initial_frame.scaleHalf();
//
//	m_frame = initial_frame;
//
//	initial_frame.saveToFile("./monoSlamOut/frame_initial_grayscale.jpg");
//
//	if (show2D && wind2D)
//		wind2D->showImage(initial_frame);
//
//
//	//	reloj.Tic();
//	CMonoSlam::InitializeNewFeature(m_FeatPatch,initial_frame);
//
//
//	 //MAAA----------------------------------------------------------
//	 std::cout<< "Number of Features In Initialization:   " << m_FeatPatch.size() << " \n";
//	 //--------------------------------------------------------------
//
//	CVectorDouble xv;
//	CQuaternionDouble q;
//
//	CMonoSlam::get_xkk(xv);
//
//	xv.extract_vector(3,q);
//	double r,p,y;
//	q.rpy(r,p,y);
//
////	reloj.Tic();
//	/******************* GROUND PLANE *******************/
//	if (show3D && wind3D)
//	{
//		{
//			opengl::CGridPlaneXYPtr obj = opengl::CGridPlaneXY::Create(-2000,2000,-2000,2000,0,100);
//			obj->setColor( 0.4f, 0.4f, 0.4f, 0.4f);
//			(world)->insert( obj );
//		}
//
//		/****************** CAMERA *******************/
//		{
//			CMatrixDouble giro(3,3), pxx(3,3),pkkg,cov3D;
//			CVectorDouble	xv;
//			CMonoSlam::get_xkk(xv);
//			CMonoSlam::get_pkk(pkkg);
//			pkkg.extractMatrix(0,0,pxx);
//			giro(0,0)=1;
//			giro(1,2)=-1;
//			giro(2,1)=1;
//
//			opengl::CEllipsoidPtr obj = opengl::CEllipsoid::Create();
//
//			giro.multiply_HCHt(pxx,cov3D);  // cov3D = (giro * pxx* ~giro)
//
//			obj->setCovMatrix(cov3D);
//			obj->setColor(0.0f,1.0f,0.0f);
//			obj->setLocation(-(float)xv[0],(float)xv[2],(float)xv[1]);
//			obj->setName("cam");
//
//			(world)->insert(obj);
//		}
//		/******************* AXIS ********************/
//		{
//			opengl::CAxisPtr obj = opengl::CAxis::Create();
//
//			obj->setColor(1.0f,0.0f,1.0f);
//			obj->setAxisLimits(-5.0f,-1.0f,-5.0f,5.0f,5.0f,5.0f);
//
//			obj->setLineWidth(5.0f);
//			obj->setFrequency( 0.5f);
//			obj->enableTickMarks(true);
//			obj->setName("axis");
//
//			(world)->insert(obj);
//		}
//		wind3D->setCameraAzimuthDeg( -100.0f );
//		wind3D->setCameraElevationDeg( 10.0f );
//		wind3D->setCameraZoom(  8.0f );
//		wind3D->setCameraPointingToPoint( 0,0,1.59f );
//
//		wind3D->unlockAccess3DScene();
//		wind3D->forceRepaint();
//
//	}
//
//	if (options.save_data_files)
//	{
//		CMatrixDouble pkk;
//		CMonoSlam::get_pkk(pkk);
//		CVectorDouble xkk;
//		CMonoSlam::get_xkk(xkk);
//		pkk.saveToTextFile("./monoSlamOut/pkkc1.txt");
//		xkk.saveToTextFile("./monoSlamOut/xkkc1.txt");
//
//		/* JL: Remove??
//
//		//MAAA-------------------------------------------------------------------------
//		observ = NULL; //sensoryframe->getObservationByClass(CLASS_ID(CObservationGPS));
//		if (observ != NULL)
//		{
//			CVectorDouble xv_store;
//			CMatrixDouble pxx_store(3,3);
//			CMonoSlam::get_xkk(xv_store);
//			xv_store = xv_store.extract_vector(0,3);
//			xv_store.saveToTextFile( format("./monoSlamOut/x%03d.txt",store_index) );
//			//pxx_store = CMonoSlam::get_pxx();
//			CMonoSlam::get_pkk(pxx_store);
//			pxx_store.setSize(3,3);
//			pxx_store.saveToTextFile(format("./monoSlamOut/p%03d.txt",store_index));
//			store_index++;
//		}
//		//------------------------------------------------------------------------------
//		*/
//	}
//}
//

