/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Perception and Robotics               |
   |      research group, University of Malaga (Spain).                        |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/core.h>
#include <mrpt/monoslam/CMonoSlam.h>
#include <mrpt/monoslam/CMonoSlamMotionModel.h>

using namespace std;
using namespace mrpt;
using namespace mrpt::utils;
using namespace mrpt::poses;
using namespace mrpt::slam;
using namespace mrpt::monoslam;
using namespace mrpt::math;
using namespace mrpt::gui;
using namespace mrpt::vision;
using namespace mrpt::hwdrivers;



/** Calculate the features' position and the Jacobian 
 *	Measurement, derivatives and texture prediction
 * \param h_Out, output. This vector contain the predicted feature position estimation.
 * \param H_Out, output. This matrix is the Jacobian of measurement prediction. H = [dh_dxv  dh_dy]
 */

void  CMonoSlam::h_calculate(CMatrixDouble &h_Out, CMatrixDouble &H_Out)
{
	// h contains the position (in undistorted pixels!??) of all features appearing in the image
	// H_Out is the matrix with the Jacobian of the measurement equation for all of the features
    // H = [dh_dxv  dh_dy]
	// hi=[u v] of pixels as the obseration vector

	// yi : Each feature has 6 parameters: (x,y,z) of a camera position,
	//                                plus (hx,hy,hz) a vector describing the direction and depth

	unsigned int numFeaturesInTheMap = (unsigned int)(m_xkk.size() - state_size) / feature_size;

 	// Size of vectors, jacobians,etc..:
	const size_t vehicle_size   = get_vehicle_size();     // 13
	const size_t feature_size   = get_feature_size();     // 6
	const size_t obs_size       = get_observation_size(); // 2 : u,v of pixels

	// Start of vehicle & landmarks within the joint state vector/matrix:
	const size_t vehicle_offset = get_vehicle_state_offset();
	const size_t feature_offset = get_feature_state_offset();
 
 	mrpt::math::CVectorDouble XfeatToPredict;	XfeatToPredict.resize(feature_size);

	//Pixel coord of observed feature
	CMatrixDouble zi(1,obs_size); 

	// Excluded band
	unsigned int excluded_band = 11; // pxls

    //Initialization h_all y  H
	h_Out.setSize(numFeaturesInTheMap,obs_size);
	h_Out.ones();
	h_Out *=-1;

	//H_Out.setSize(0,0);
	H_Out.setSize(numFeaturesInTheMap*obs_size,m_xkk.size());
	H_Out.zeros();

    for (unsigned int i = 0; i<numFeaturesInTheMap ; i++)
	{
        m_xkk.extract_vector(state_size+i*feature_size,XfeatToPredict);

        // Measurement equation
		// Returns the pixel cooordinates (distorted) onto we expect to see this feature
		// Performs a Projection from 3D (yi) to 2D (u,v) 
		hi( XfeatToPredict, zi,i );
		//H_Out.setSize(H_Out.getRowCount()+2,m_xkk.size());		//Se redimensiona pero conserva los valores
		
	    // Is visible and predicted in the image?
		if (( zi(0,0)>excluded_band ) && ( zi(0,0)<cam.cam_ncols()-excluded_band ) && ( zi(0,1)>excluded_band ) && ( zi(0,1)<cam.cam_nrows()-excluded_band ))
		{
			h_Out.insertMatrix(i,0,zi);

            //Calculate derivatives (2 ecuaciones linealizadas para cada caracteristica)
			// Hi is the matrix with the Jacobian of the measurement equation for each of those features
			// Hi = [dhi_dxv  dhi_dy1 ..  dhi_dyn  ],hi=[u,v]
			CMatrixDouble Hi(obs_size,vehicle_size+feature_size);
 			calculate_Hi(XfeatToPredict, zi, i, Hi);

			//H_Out.setSize(H_Out.getRowCount()+2,m_xkk.size());		//Se redimensiona pero conserva los valores
			//H_Out.insertMatrix(H_Out.getRowCount()-2,0,Hi);

			H_Out.insertMatrix(i*obs_size,0,Hi);

		}
		else{
			std::cout<<" The feature #" << i << " is out of the band and not visible! \n";
			//std::cin.get();
		}

	}
}


/******************************************************************************************************************************/
/*
	 Hi is the matrix with the Jacobian of the measurement equation for each of those features
     Hi = [dhi_dxv  dhi_dy1 ..  dhi_dyn  ]
	 h  = [u,v] projected and distorted pixel coordninates
*/

 void  CMonoSlam::calculate_Hi(mrpt::math::CVectorDouble &yi, CMatrixDouble &zi, unsigned int i,CMatrixDouble &Hi)
{
	CMatrixDouble matrix_dh_dxv(2,state_size);
	CMatrixDouble matrix_dh_dyi(2,feature_size);
 
	dh_dxv(yi,zi,matrix_dh_dxv);//Hx
	dh_dyi(yi,zi,matrix_dh_dyi);//Hy

	//Hi = [dh_dxv   0 .. dh_dyi .. 0];
	Hi.insertMatrix(0,0,matrix_dh_dxv);
	Hi.insertMatrix(0,state_size/*+feature_size*i*/,matrix_dh_dyi);

	//std::cout<<"[Mono-Slam] matrix_dh_dyi : "<< i << " \n " << matrix_dh_dyi<<"\n";

  	//Hi.saveToTextFile("./monoSlamOut/Hi.txt");

}

void  CMonoSlam::dh_dxv(mrpt::math::CVectorDouble &yi, CMatrixDouble &zi, CMatrixDouble &m_dh_dxv)
{
	CMatrixDouble matrix_dh_drw(2,3);
	CMatrixDouble matrix_dh_dqwr(2,4);

	dh_drw(yi,zi,matrix_dh_drw);
    dh_dqwr(yi,zi,matrix_dh_dqwr);
	m_dh_dxv.insertMatrix(0,0,matrix_dh_drw);
	m_dh_dxv.insertMatrix(0,3,matrix_dh_dqwr);
}


void  CMonoSlam::dh_dyi(mrpt::math::CVectorDouble &yi, CMatrixDouble &zi, CMatrixDouble &m_dh_dy)
{
	//it means that m_dh_dyi is computed

	CMatrixDouble matrix_dh_dhrl(2,3);
	CMatrixDouble matrix_dhrl_dyi(3,feature_size);

    dh_dhrl(yi,zi,matrix_dh_dhrl);

    dhrl_dy(yi,matrix_dhrl_dyi);

	m_dh_dy = matrix_dh_dhrl * matrix_dhrl_dyi;

  	//std::cout<<"[Mono-Slam] dhrl_dyi: \n "<< matrix_dhrl_dyi << "\n";
 	//std::cout<<"[Mono-Slam] dh_dhrl: \n "<< matrix_dh_dhrl << "\n";

}
/******************************************************************************************************************************/
/******************************************************************************************************************************/

void  CMonoSlam::dh_drw(mrpt::math::CVectorDouble &yi, CMatrixDouble &zi, CMatrixDouble &m_dh_drw)
{
	CMatrixDouble matrix_dh_dhrl(2,3);
	CMatrixDouble matrix_dhrl_drw(3,3);

	dh_dhrl(yi,zi,matrix_dh_dhrl);
    dhrl_drw(yi,matrix_dhrl_drw);
	m_dh_drw = matrix_dh_dhrl * matrix_dhrl_drw;
}

void  CMonoSlam::dh_dqwr(mrpt::math::CVectorDouble &yi, CMatrixDouble &zi, CMatrixDouble &m_dh_dqwr)
{
	CMatrixDouble matrix_dh_dhrl(2,3);
	CMatrixDouble matrix_dhrl_dqwr(3,4);

	dh_dhrl(yi,zi,matrix_dh_dhrl);

    dhrl_dqwr(yi,matrix_dhrl_dqwr);
	m_dh_dqwr = matrix_dh_dhrl * matrix_dhrl_dqwr;
}
/******************************************************************************************************************************/

void  CMonoSlam::dh_dhrl(mrpt::math::CVectorDouble &yi, CMatrixDouble &zi, CMatrixDouble &dh_dhrl)
{
	CMatrixDouble matrix_dhd_dhu(2,2);
	CMatrixDouble matrix_dhu_dhrl(2,3);

	//dhd_dhu(zi,matrix_dhd_dhu);
 //   dhu_dhrl(yi,matrix_dhu_dhrl);
	//m_dh_dhrl = matrix_dhd_dhu * matrix_dhu_dhrl;

	//MAAA: testing new method
	mrpt::math::TPoint3D hc(m_hrl_i[0] ,m_hrl_i[1] ,m_hrl_i[2] );
    cam.jacobian_project_with_distortion(hc ,dh_dhrl );

}

/******************************************************************************************************************************/

void  CMonoSlam::dhrl_dy(mrpt::math::CVectorDouble &yi, CMatrixDouble &m_dhrl_dy)
{
	mrpt::math::CVectorDouble rw,yi_temp,dhrl_dthetai,dhrl_dphii,dhrl_drhoi;
	rw.resize(3);	yi_temp.resize(3);	dhrl_dthetai.resize(3);	dhrl_dphii.resize(3);	dhrl_drhoi.resize(3);

	CMatrixDouble Rrw(3,3),Rwr(3,3),m_dhrl_dthetai(3,1),m_dhrl_dphii(3,1),m_dhrl_drhoi(3,1), m_dhrl_dyi(3,3);
	CQuaternionDouble q;

	m_xkk.extract_vector(0,rw);
	m_xkk.extract_vector(3,q);
	yi.extract_vector(0,yi_temp);
	q.q_rotation_matrix(Rwr);
	Rrw=Rwr.inv();
	float theta(yi[3]);
	float phi(yi[4]);
	float rho(yi[5]);

	dhrl_dthetai[0] = cos(phi) * cos(theta);
	dhrl_dthetai[2] = -cos(phi) * sin(theta);
	m_dhrl_dthetai = Rrw * ~dhrl_dthetai;

	dhrl_dphii[0] = -sin(phi) * sin(theta);
	dhrl_dphii[1] = -cos(phi);
	dhrl_dphii[2] = -sin(phi) * cos(theta);
	m_dhrl_dphii = Rrw * ~dhrl_dphii;

	m_dhrl_drhoi = ~yi_temp - ~rw;
	m_dhrl_drhoi = Rrw * m_dhrl_drhoi;

	m_dhrl_dyi = Rrw;
	m_dhrl_dyi *=rho;

	m_dhrl_dy.insertMatrix(0,0,m_dhrl_dyi);
	m_dhrl_dy.insertMatrix(0,3,m_dhrl_dthetai);
	m_dhrl_dy.insertMatrix(0,4,m_dhrl_dphii);
	m_dhrl_dy.insertMatrix(0,5,m_dhrl_drhoi);
}


/******************************************************************************************************************************/

void  CMonoSlam::dhrl_drw(mrpt::math::CVectorDouble &yi, CMatrixDouble &m_dhrl_drw)
{
	CMatrixDouble temp(3,3);
	CQuaternionDouble q;

	m_xkk.extract_vector(3,q);
	q.q_rotation_matrix(temp);

	m_dhrl_drw = temp.inv();
	m_dhrl_drw *= -1.0f * yi[5];
}
/******************************************************************************************************************************/

void  CMonoSlam::dhd_dhu(CMatrixDouble &zi, CMatrixDouble &m_dhd_dhu)
{
	CMatrixDouble a_inv(2,2); //,a(2,2);

	// JL: TODO CHECK!!!!!
	cam.jacob_undistor_fm( TPixelCoordf(zi(0,0),zi(0,1)), a_inv);
		
    //std::cout <<"a_inv : \n" << a_inv << "\n";
	m_dhd_dhu = a_inv.inv();;



}
/******************************************************************************************************************************/

void  CMonoSlam::dhu_dhrl(mrpt::math::CVectorDouble &yi, CMatrixDouble &m_dhu_dhrl)
{
	CMatrixDouble Rrw(3,3),Rwr(3,3);
	CQuaternionDouble q;
	m_xkk.extract_vector(3,q);
	q.q_rotation_matrix(Rwr);
	Rrw = Rwr.inv();

	float theta(yi[3]);
	float phi(yi[4]);
	float rho(yi[5]);
	mrpt::math::CVectorDouble mi;
	mi.resize(3);
	mi[0]= cos(phi) * sin(theta);
	mi[1]= -sin(phi);
	mi[2]= cos(phi) * cos(theta);

	CMatrixDouble temp(3,1),hc(3,1);
	mrpt::math::CVectorDouble yi_temp,rw;
	yi_temp.resize(3);
	rw.resize(3);
	yi.extract_vector(0,yi_temp);
	m_xkk.extract_vector(0,rw);
	temp = ~yi_temp - ~rw;
	temp *= rho;
	temp = temp + ~mi;
	hc = Rrw * temp;

	m_dhu_dhrl(0,0) = -cam.cam_fx()/hc(2,0);
	m_dhu_dhrl(0,2) = hc(0,0)* cam.cam_fx()/square(hc(2,0));
	m_dhu_dhrl(1,1) = -cam.cam_fx()/hc(2,0);
	m_dhu_dhrl(1,2) = hc(1,0)* cam.cam_fx()/square(hc(2,0));
}

/******************************************************************************************************************************/

void  CMonoSlam::dhrl_dqwr(mrpt::math::CVectorDouble &yi, CMatrixDouble &m_dhrl_dqwr)
{
	mrpt::math::CVectorDouble rw,mi,yi_temp;
	rw.resize(3);
	mi.resize(3);
	yi_temp.resize(3);
	CQuaternionDouble q;
	m_xkk.extract_vector(0,rw);
	m_xkk.extract_vector(3,q);
	float theta(yi[3]);
	float phi(yi[4]);
	float rho(yi[5]);

	mi[0] = cos(phi) * sin(theta);
	mi[1] = -sin(phi);
	mi[2] = cos(phi) * cos(theta);

	CMatrixDouble temp(3,1);
	yi.extract_vector(0,yi_temp);
	temp = ~yi_temp - ~rw;

	temp *=rho;
	temp += ~mi;

	mrpt::math::CQuaternion<double>		auxQ(q.q_conj());

	m_dhrl_dqwr = dRq_times_a_by_dq(auxQ,temp);
	m_dhrl_dqwr = m_dhrl_dqwr * dqbar_by_dq();
}

/*************************************************NEWBITS*******************************************************/
/** Jacobians of quaternion product with respect to arguments. Assume product is q3 = q2 * q1
* It happens that dq3_by_dq1 only depends on q2 and dq3_by_dq2 on q1
*/
/*void dq3_by_dq1(mrpt::math::CQuaternion<double>& q2,CMatrixDouble &M)	//matrix(4x4)
{
  //M.setSize(4,4);
  M(0,0)=q2.r();	M(0,1)=-q2.x();		M(0,2)=-q2.y();		M(0,3)=-q2.z();
  M(1,0)=q2.x();	M(1,1)=q2.r();		M(1,2)=-q2.z();		M(1,3)=q2.y();
  M(2,0)=q2.y();	M(2,1)=q2.z();		M(2,2)=q2.r();		M(2,3)=-q2.x();
  M(3,0)=q2.z();	M(3,1)=-q2.y();		M(3,2)=q2.x();		M(3,3)=q2.r();
}


void dq3_by_dq2(mrpt::math::CQuaternion<double>& q1,CMatrixDouble &M)	//matrix(4x4)
{
  //M.setSize(4,4);
  M(0,0)=q1.r();	M(0,1)=-q1.x();		M(0,2)=-q1.y();		M(0,3)=-q1.z();
  M(1,0)=q1.x();	M(1,1)=q1.r();		M(1,2)=q1.z();		M(1,3)=-q1.y();
  M(2,0)=q1.y();	M(2,1)=-q1.z();		M(2,2)=q1.r();		M(2,3)=q1.x();
  M(3,0)=q1.z();	M(3,1)=q1.y();		M(3,2)=-q1.x();		M(3,3)=q1.r();
}*/

/** Jacobian of quaternion inverse: this is a constant matrix that doesn't
* depend on the quaternion
*/
CMatrixDouble  CMonoSlam::dqbar_by_dq()	//matrix(4x4)
{
  CMatrixDouble M(4,4);
  M(0,0)=1;
  M(1,1)=-1;
  M(2,2)=-1;
  M(3,3)=-1;

  return M;
}

/******************************************************************************************************************************/
/** Jacobian for rotation derived from quaternion multiplied by a vector
* This is tricky because really we need a tensor to do it neatly I think!
* But do it here without
* Return 3x4 matrix
*/
/** Component Jacobians: each returns a 3x3 matrix which is the derivative
* of the rotation matrix derived from a quaternion w.r.t. each quaternion
* element
*/
CMatrixDouble  CMonoSlam::dR_by_dq0(CQuaternionDouble &q)	//matrix(3x3)
{

  CMatrixDouble M(3,3);
  M(0,0)=2*q.r();	M(0,1)=-2*q.z();	M(0,2)=2*q.y();
  M(1,0)=2*q.z();	M(1,1)=2*q.r();		M(1,2)=-2*q.x();
  M(2,0)=-2*q.y();	M(2,1)=2*q.x();		M(2,2)=2*q.r();

  return M;
}
/******************************************************************************************************************************/

CMatrixDouble  CMonoSlam::dR_by_dqx(CQuaternionDouble &q)	//matrix(3x3)
{

  CMatrixDouble M(3,3);
  M(0,0)=2*q.x();	M(0,1)=2*q.y();		M(0,2)=2*q.z();
  M(1,0)=2*q.y();	M(1,1)=-2*q.x();	M(1,2)=-2*q.r();
  M(2,0)=2*q.z();	M(2,1)=2*q.r();		M(2,2)=-2*q.x();

  return M;
}
/******************************************************************************************************************************/

CMatrixDouble  CMonoSlam::dR_by_dqy(CQuaternionDouble &q)	//matrix(3x3)
{

  CMatrixDouble M(3,3);
  M(0,0)=-2*q.y();	M(0,1)=2*q.x();		M(0,2)=2*q.r();
  M(1,0)=2*q.x();	M(1,1)=2*q.y();		M(1,2)=2*q.z();
  M(2,0)=-2*q.r();	M(2,1)=2*q.z();		M(2,2)=-2*q.y();
  return M;
}
/******************************************************************************************************************************/
CMatrixDouble  CMonoSlam::dR_by_dqz(CQuaternionDouble &q)	//matrix(3x3)
{
  CMatrixDouble M(3,3);
  M(0,0)=-2*q.z();	M(0,1)=-2*q.r();	M(0,2)=2*q.x();
  M(1,0)=2*q.r();	M(1,1)=-2*q.z();	M(1,2)=2*q.y();
  M(2,0)=2*q.x();	M(2,1)=2*q.y();		M(2,2)=2*q.z();
  return M;
}
/******************************************************************************************************************************/
/** Normalising Jacobian
*/
CMatrixDouble  CMonoSlam::dqnorm_by_dq(CQuaternionDouble &q)	//matrix (4x4)
{
  float qq = q.r()*q.r() + q.x()*q.x() + q.y()*q.y() + q.z()*q.z();

  CMatrixDouble M(4,4);
  M(0,0)=dqi_by_dqi(q.r(), qq);
  M(0,1)=dqi_by_dqj(q.r(), q.x(), qq);
  M(0,2)=dqi_by_dqj(q.r(), q.y(), qq);
  M(0,3)=dqi_by_dqj(q.r(), q.z(), qq);

  M(1,0)=dqi_by_dqj(q.x(), q.r(), qq);
  M(1,1)=dqi_by_dqi(q.x(), qq);
  M(1,2)=dqi_by_dqj(q.x(), q.y(), qq);
  M(1,3)=dqi_by_dqj(q.x(), q.z(), qq);

  M(2,0)=dqi_by_dqj(q.y(), q.r(), qq);
  M(2,1)=dqi_by_dqj(q.y(), q.x(), qq);
  M(2,2)=dqi_by_dqi(q.y(), qq);
  M(2,3)=dqi_by_dqj(q.y(), q.z(), qq);

  M(3,0)=dqi_by_dqj(q.z(), q.r(), qq);
  M(3,1)=dqi_by_dqj(q.z(), q.x(), qq);
  M(3,2)=dqi_by_dqj(q.z(), q.y(), qq);
  M(3,3)=dqi_by_dqi(q.z(), qq);

  return M;
}
/******************************************************************************************************************************/
/** Normalising a 3D vector
*/
CMatrixDouble  CMonoSlam::dvnorm_by_dv(mrpt::math::CVectorDouble &v)	//matrix(3x3)
{
  float vv = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];

  // We can use dqi_by_dqi and dqi_by_dqj functions because they are the same
  CMatrixDouble M(3,3);
  M(0,0)=dqi_by_dqi(v[0], vv);
  M(0,1)=dqi_by_dqj(v[0], v[1], vv);
  M(0,2)=dqi_by_dqj(v[0], v[2], vv);

  M(1,0)=dqi_by_dqj(v[1], v[0], vv);
  M(1,1)=dqi_by_dqi(v[1], vv);
  M(1,2)=dqi_by_dqj(v[1], v[2], vv);

  M(2,0)=dqi_by_dqj(v[2], v[0], vv);
  M(2,1)=dqi_by_dqj(v[2], v[2], vv);
  M(2,2)=dqi_by_dqi(v[2], vv);

  return M;
}
/******************************************************************************************************************************/
CMatrixDouble  CMonoSlam::dRq_times_a_by_dq (CQuaternionDouble &q, CMatrixDouble &aMat) //matrix(3x4)
//q --> es el quaternion del vector de estado
//a --> el rayo que une el centro de la camara con la caracteristica (hLhatRi) "desproyectado", normalizado PERO SIN ROTAR
{
  CMatrixDouble TempR(3,3);
  CMatrixDouble Temp31(3,1);
  CMatrixDouble dRq_times_a_by_dq(3,4);

  // Make Jacobian by stacking four vectors together side by side
  TempR = dR_by_dq0(q);
  Temp31 = TempR * aMat;
  dRq_times_a_by_dq.insertMatrix(0,0,Temp31);

  TempR = dR_by_dqx(q);
  Temp31 = TempR * aMat;
  dRq_times_a_by_dq.insertMatrix(0,1,Temp31);

  TempR = dR_by_dqy(q);
  Temp31 = TempR * aMat;
  dRq_times_a_by_dq.insertMatrix(0,2,Temp31);

  TempR = dR_by_dqz(q);
  Temp31 = TempR * aMat;
  dRq_times_a_by_dq.insertMatrix(0,3,Temp31);

  return dRq_times_a_by_dq;
}
/******************************************************************************************************************************/
/** Auxiliary functions used by dqnorm_by_dq()
* Value of diagonal element of Jacobian
*/
float  CMonoSlam::dqi_by_dqi(float qi, float qq)
{
  return (1 - qi*qi / (qq*qq)) / qq;
}
/******************************************************************************************************************************/
/** Value of off-diagonal element of Jacobian
*/
float  CMonoSlam::dqi_by_dqj(float qi, float qj, float qq)
{
  return -qi * qj / (qq*qq*qq);
}

