/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Perception and Robotics               |
   |      research group, University of Malaga (Spain).                        |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/core.h>
#include <mrpt/monoslam/CMonoSlam.h>
#include <mrpt/monoslam/CMonoSlamMotionModel.h>

using namespace std;
using namespace mrpt;
using namespace mrpt::utils;
using namespace mrpt::poses;
using namespace mrpt::slam;
using namespace mrpt::monoslam;
using namespace mrpt::math;
using namespace mrpt::gui;
using namespace mrpt::vision;
using namespace mrpt::hwdrivers;





//----------------------------------------------------------------
/*  Main Loop of the CMonoSlamInterface */
//----------------------------------------------------------------
void  CMonoSlam::go_one_step( const CActionCollectionPtr &optional_odometry )
{
	MRPT_TRY_START;
	ASSERTMSG_(m_camera.present(),"No source of video prepared in m_camera!!")

	++m_iterations_count;  // For debugging/logging purposes only

	CMatrixDouble u(1,1);

	//-------------------------------------------------------------------------
	// GET OBSERVATION
	//-------------------------------------------------------------------------
	// Grab the next image from the user source of video:
	m_camera->doProcess();

	// Recover grabbed images:
	{
		CObservationImagePtr	observation;

		CGenericSensor::TListObservations  lstCamObs;
		m_camera->getObservations( lstCamObs );
		if (lstCamObs.empty())
			THROW_EXCEPTION("Error: No image grabbed from specified camera.")
		ASSERT_( IS_CLASS( lstCamObs.begin()->second, CObservationImage ) );

		// Get image and save it in "m_frame"
		observation = CObservationImagePtr(lstCamObs.begin()->second);
		this->m_frame.copyFastFrom( observation->image );

		// Assert image is grayscale:
		this->m_frame.grayscaleInPlace();

		//Set delta_t between frames:
		ASSERT_(observation->timestamp!=INVALID_TIMESTAMP)
		if (m_timeLastFrame==INVALID_TIMESTAMP)
				mmm.set_delta_t( 1.0/15.0 );
		else	mmm.set_delta_t( system::timeDifference(m_timeLastFrame,observation->timestamp) );

		m_timeLastFrame = observation->timestamp;	// For the next step

		if (options.verbose)
			std::cout<< "[mono-slam] Time since last frame: "<<mmm.get_delta_t()<<"\n";
	}

	// save the actions, if given by the user:
	// ----------------------------------------------
	m_action = optional_odometry;



	//std::cout<<"[mono-slam] Frame: "<<m_framecolor.getHeight()<<","<<m_framecolor.getWidth()<< endl;

	/******************************SAVE LOGS*******************************/
	if (options.save_data_files)
	{
		m_pkk.saveToTextFile(format("./monoSlamOut/p%05d.txt",m_iterations_count));
		m_xkk.saveToTextFile(format("./monoSlamOut/x%05d.txt",m_iterations_count));
	}

	//-------------------------------------------------------------------------
 	//   Here's the meat! Run the Kalman Filter iteration
 	//-------------------------------------------------------------------------
	CMonoSlam::KF_options.method  = mrpt::bayes::kfEKFNaive; // JL: just for debuggin! !!!
	CMonoSlam::KF_options.verbose = options.verbose;
 
	if (options.verbose)
	{
		static bool first_show_KFopts = true;
		if (first_show_KFopts)
		{
			first_show_KFopts = false;
			cout << "[mono-slam] Calling KF method with these parameters:" << endl;
			KF_options.dumpToConsole();
			cout << endl;
		}
	}

	// Run One Kalman Iteration
	//   If this is the first step, initialize the filter with a special function:
	ASSERT_(m_xkk.size() >= get_vehicle_size());

	if ( m_xkk.size() == get_vehicle_size()  )
	{
		CMonoSlam::InitializeNewFeature(m_FeatPatch,m_frame);
		//MAAA 
		std::cout<< "Number of Features In Initialization:   " << m_FeatPatch.size() << " \n";
	}
	else
	{
		// Normal KF iteration:
		ASSERT_(m_h.getRowCount()!=0 || m_zz.getRowCount()!=0);  // JL: ????

		// JL says: TODO: manage data_association as expected in KF so the bool parameter can be removed...
		CMonoSlam::runOneKalmanIteration(); //true); // <- this was a temporary hack...
	}

 	//CMatrixDouble H ,h,S;
	//if (type == 3)
	//	hkalman.prediction_state_and_measure(u,H,S,h);
	//else if (type == 2)
	//{
	//	ukalman.StatePrediction(u);
	//	ukalman.PredictedMeasurements(h,S);
	//}
	//else if (type == 1)
	//	ikalman.prediction_state_and_measure(u,H,S,h);
	//else
	//	kalman.prediction_state_and_measure(u,H,S,h);

	//std::cout<<"update\n";
	//if (type==3)
	//	hkalman.update(H,RR,h,zz,match);
	//else if (type==2)
	//	ukalman.Update(match,RR,zz);
	//else if (type==1)
	//	ikalman.update(H,RR,h,zz,match);
	//else
	//	kalman.update(HH,RR,hh,zz);
 	//std::cout<<"update : "<<reloj.Tac()<<"\n";


	/***********************************************************************************************************/
	// JL: Oh my! this section SHOULD NOT be needed someday!!

	//------------------------------------------------------------------------
	// This section of code is a workaround for a situation where
    // the camera suddenly shoots off at high speed, which must be
    // a bug perhaps in the state update.  If the camera suddenly accelerates
    // at a high rate then this just sets the state back to the previous one.
	//------------------------------------------------------------------------

#if 1

 	CVectorDouble velocity,v2;
	velocity.resize(3);
	v2.resize(3);
	m_xkk.extract_vector(0,velocity);
	//std::cout<<"xv actual: "<<velocity<<"\n";
	prev_xv.extract_vector(0,v2);
	//std::cout<<"xv anterior: "<<v2<<"\n";
	velocity = (velocity - v2) / u(0,0);			//u(0,0) = delta_t
	float speed = sqrt(velocity[0]*velocity[0]+velocity[1]*velocity[1]+velocity[2]*velocity[2]);
	//std::cout<<"speed: "<<speed<<"	prev_speed: "<<prev_speed<<"\n"<<"	velocity: "<<velocity<<"\n";

	if (prev_speed == 0)
		prev_speed = speed;
	else
	{
		float speed_change = speed / prev_speed;
		//std::cout<<"speed_change: "<<speed_change<<"\n";
		int temp=0;
		for (unsigned int i=0; i<prev_xv.size();i++)	//es el primero
			if (prev_xv[i]!=0) temp++;

		//std::cout<<"temp: "<<temp<<"\n";
		if ((speed > 0.7) && (speed_change > 1.2) && (temp!=0))	//cambio muy brusco
		{
			std::cout<<"Velocity sudden increase!\n";
			std::cout<<"speed: "<<speed<<"	velocity: "<<velocity<<"\n";//std::cin.get();
			m_xkk.insert_vector(0,prev_xv);
 			//std::cin.get();
		}
		else
		{
			m_xkk.extract_vector(0,prev_xv);
		}
		prev_speed = speed;
	}
	/***********************************************************************************************************/
#endif

	//------------------------------------------------------------------------
	// SHOW RESULTS
	//------------------------------------------------------------------------

	if (1) // Prepare progress views in 2D & 3D:
	{
		static CTicTac	tim;

		tim.Tic();

		CMatrixDouble XYZ = CMonoSlam::XYZ6D_TO_XYZ3D();
		XYZ.setSize(XYZ.getRowCount(),XYZ.getColCount()+1);

		// The added right-most column is for avoiding recomputing everything in the 3D representation:
		// XYZ(x coord, y coord, z coord, type)
		// type = 0 --> No visible
		// type = 1 --> Visible and Matching
		// type = 2 --> Visible and no Matching

		// Create an auxiliary version of the frame in color:
		CImage  temp_img_color;
		m_frame.colorImage(temp_img_color); 
		plot_results_2D(m_zz,m_hh,m_h,m_S,XYZ,temp_img_color, KF_options.method,m_match,options.verbose,options.save_data_files);

		// Prepare 3D view:
		plot_results_3D(XYZ);

		if (options.verbose)
			cout << "[go_one_step] 2D&3D views preparation: " << tim.Tac()*1e3 << " ms" << endl;
	}

	//------------------------------------------------------------------------
	// ERASE BAD FEATURES
	//------------------------------------------------------------------------

	bool bDeletedAfeature=false;

#if 0

	//reloj.Tic();
	CVectorDouble	x;
	CMatrixDouble	p;
	CMonoSlam::get_xkk(x);
	CMonoSlam::get_pkk(p);
	if (options.verbose)
		std::cout<< "[mono-slam] Selecting features to be removed." << endl;

 

 	for (size_t t = m_FeatPatch.size() ; t > 0 ; t--)
	{
		if(bDeletedAfeature)
			continue;

		if (m_FeatPatch.get_hits(t-1)<=0)
		{
			if (options.verbose)
				std::cout<<"******************* Elimination of Feature: "<<m_FeatPatch.get_name(t-1)<<" *******************\n";
	
			//std::cout<<" m_Hx_Hy size :" << m_Hx_Hy.getRowCount() <<"\n";
			//std::cout<<" m_Hx_Hy[0,0] :" << m_Hx_Hy(0,0) <<"\n";
			//std::cout<<" m_Hx_Hy[2,0] :" << m_Hx_Hy(2,0) <<"\n";
			//std::cout<<" m_Hx_Hy[4,0] :" << m_Hx_Hy(4,0) <<"\n";

			if (m_show2D && m_show3D && wind3D)
			{
            	world = wind3D->get3DSceneAndLock();
				//for (k = 2*(t-1)+3; k < (world)->classinfo->objectSize-2; k++)
				//{
				//	(world)->[k]=(*world)->m_objects[k+2];
				//}
				//(*world)->m_objects.resize((*world)->m_objects.size()-2);

 			    (world)->removeObject((world)->getByName( format("E%02d",m_FeatPatch.get_name(t-1)) ));
				(world)->removeObject((world)->getByName( format("S%02d",m_FeatPatch.get_name(t-1)) ));

				wind3D->unlockAccess3DScene();

				// Update window:
				wind3D->forceRepaint();
			}

			size_t ObsSize = get_observation_size();
			m_FeatPatch.delete_feature(t-1);

			m_featList.erase(m_featList.begin()+t-1);
			//const size_t &index = t-1;
			//{
			//	if (index>=m_featList.size())
			//		THROW_EXCEPTION("Feature for delete out of bounds in CFeaturePatch::delete_feature");
			//	for (size_t i=index;i<m_featList.size()-1;i++)
			//	{
			//		m_featList[i]=m_featList[i+1];
			//	}
			//	m_featList.resize(m_featList.size()-1);
			//}

    		CMonoSlam::delete_feature(t-1);

			bDeletedAfeature = true;

			//Update m_h and m_Hx_Hy
            function_h( m_h, m_Hx_Hy);

			//std::cout<<" m_Hx_Hy size :" << m_Hx_Hy.getRowCount() <<"\n";
			//std::cout<<" m_Hx_Hy[0,0] :" << m_Hx_Hy(0,0) <<"\n";
			//std::cout<<" m_Hx_Hy[2,0] :" << m_Hx_Hy(2,0) <<"\n";
			//std::cout<<" m_Hx_Hy[4,0] :" << m_Hx_Hy(4,0) <<"\n";

			std::cout<< "[mono-slam] ******************* feature #" << t-1 << " removed. ******************" << endl;


			std::cin.get();

		}
	}
 
	//std::cout<<"erase features: "<<reloj.Tac()<<"\n";
	//--------------------------------------------------------------------------
#endif 

	//---------------------------------------------------------------------------
	//SEARCH MORE FEATURES
	//---------------------------------------------------------------------------

	unsigned int num_observed_feat  = (unsigned int)(m_zz.getRowCount()/2);
	unsigned int num_Predicted_feat = (unsigned int)(m_hh.getRowCount()/2);
	unsigned int num_feat_total     = (unsigned int)(m_h.getRowCount());
	unsigned int num_feat_          = (unsigned int)(m_featList.size());
	unsigned int num_Landmark_total = (unsigned int)(m_FeatPatch.size());

	std::cout<< "\n";
	std::cout<<" num Predicted feat :" << num_Predicted_feat <<"\n";
	std::cout<<" num Matched feat   :" << num_observed_feat  <<"\n";
	std::cout<< "\n";


#if 0	
	
	if( /*!bDeletedAfeature && */num_observed_feat <=   ) 
	{
	//reloj.Tic();

		std::cout<< "-------------------------Before adding a feature-------------------------\n";
		std::cout<<" m_h.size           :" << num_feat_total     <<"\n";
		std::cout<<" m_featList.size    :" << num_feat_          <<"\n";
		std::cout<<" m_FeatPatch.size   :" << num_Landmark_total <<"\n";
		std::cout<< "\n";

		//if (options.verbose)
			std::cout<<"Searching More Features started!\n";
			//std::cin.get();

		  CMatrixDouble  h = m_h;
 		  CMonoSlam::SearchMoreFeature(m_FeatPatch,m_frame,h);

 		////Update m_S
		////-----------------------
		function_h( m_h, m_Hx_Hy);
	
		CMatrixDouble Ri;
		Ri.unit(2);
		Ri *= square( mmm.get_sigma_image_noise() ); 
		//Ri += 20; 

  		const size_t obs_size  = get_observation_size();
		m_Hx_Hy.multiply_HCHt(m_pkk,m_S);

		//std::cout<<"[Mono-SLAM] m_S: "<<m_S<<"\n";
		const size_t num_feats_in_map = (m_xkk.size() - state_size) / feature_size;
 		for (size_t i=0;i<num_feats_in_map;i++)
		{					
			size_t  obs_idx_off = i*obs_size;
			if(m_h(i,0)>0 && m_h(i,1)>0)
			{
				for (size_t j=0;j<obs_size;j++)
					for (size_t k=0;k<obs_size;k++)
						m_S(obs_idx_off+j,obs_idx_off+k) += Ri(j,k);
 			}
			else
			{
				for (size_t j=0;j<obs_size;j++)
					m_S(obs_idx_off+j,obs_idx_off+j) += (KFTYPE) KF_options.veryLargeR2;
			}
		}
 
		//if (options.verbose)
			std::cout<<"Searching More Features finished.\n";
 		

 		num_feat_total     = (unsigned int)(m_h.getRowCount());
		num_feat_          = (unsigned int)(m_featList.size());
		num_Landmark_total = (unsigned int)(m_FeatPatch.size());

		std::cout<< "-------------------------After adding a feature-------------------------------\n";
		std::cout<<" m_h.size           :" << num_feat_total     <<"\n";
		std::cout<<" m_featList.size    :" << num_feat_          <<"\n";
		std::cout<<" m_FeatPatch.size   :" << num_Landmark_total <<"\n";
 		std::cout<< "\n";
			
		std::cout<< "Press any Key ...\n";std::cin.get();
	}


#endif 

	MRPT_TRY_END;

}

