/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Perception and Robotics               |
   |      research group, University of Malaga (Spain).                        |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/core.h>
#include <mrpt/monoslam/CMonoSlam.h>
#include <mrpt/monoslam/CMonoSlamMotionModel.h>

using namespace std;
using namespace mrpt;
using namespace mrpt::utils;
using namespace mrpt::poses;
using namespace mrpt::slam;
using namespace mrpt::monoslam;
using namespace mrpt::math;
using namespace mrpt::gui;
using namespace mrpt::vision;
using namespace mrpt::hwdrivers;


/** Load all the parameters for Mono-SLAM, Kalman-Filter, and the motion model.
  *  The configuration file (.ini-like file) may have these sections:
  *
  * \sa saveParameters
  */
void  CMonoSlam::loadParameters( const mrpt::utils::CConfigFileBase &source)
{
	// MONOSLAM
	options.loadFromConfigFile(source,"MONOSLAM");

	// KALMAN
	KF_options.loadFromConfigFile(source,"KALMAN");

	// MOTION_MODEL
	mmm.sigma_lin_vel	= source.read_double("MOTION_MODEL","sigma_lin_vel", mmm.sigma_lin_vel, true);
	mmm.sigma_ang_vel	= source.read_double("MOTION_MODEL","sigma_ang_vel", mmm.sigma_ang_vel, true);

	// CAMERA_MODEL
	cam.loadFromConfigFile(source,"CAMERA_MODEL");
}

/** Save all the parameters for Mono-SLAM, Kalman-Filter, and the motion model.
  *
  * \sa loadParameters
  */
void  CMonoSlam::saveParameters( mrpt::utils::CConfigFileBase &config)
{
	THROW_EXCEPTION("TO DO")
}


/** Initilization of default parameters */
CMonoSlam::TOptions::TOptions() :
	verbose				( true ),
	save_data_files		( false ),
	match_threshold_up	(0.85),
	match_threshold_down(0.72),
	threshold_ini		(0.0035),
	threshold_new		(0.001),
	radius				(40),		//!< Minimum radius between detected image features
	sigma				(7.0),		//!< Sigma value in the Harris method
#ifdef _DEBUG
	max_num_feat_to_detect(3),	//!< max num of features to detect
#else
	max_num_feat_to_detect(14),	//!< max num of features to detect
#endif
	lambda_init			(1.0/0.8),		//!< Initial Inverse Depth estimation the first time a feature is observed
	std_lambda			(0.25),			//!< Inverse Depth sigma for the first time a feature is observed
	sigma_image_noise	(1.0)			//!< image noise in pixels
{
}

/** See utils::CLoadableOptions */
void  CMonoSlam::TOptions::loadFromConfigFile( const mrpt::utils::CConfigFileBase &source, const std::string &section)
{
	MRPT_LOAD_CONFIG_VAR(verbose,bool,				source,section)
	MRPT_LOAD_CONFIG_VAR(save_data_files,bool,		source,section)
	MRPT_LOAD_CONFIG_VAR(match_threshold_up,double,	source,section)

	// ...

}

/** See utils::CLoadableOptions  */
void  CMonoSlam::TOptions::dumpToTextStream(CStream  &out) const
{
	out.printf("\n----------- [CMonoSlam::TOptions] ------------ \n\n");

	LOADABLEOPTS_DUMP_VAR(verbose, bool)
	LOADABLEOPTS_DUMP_VAR(save_data_files, bool)
	LOADABLEOPTS_DUMP_VAR(match_threshold_up, double)
	// ...

	out.printf("\n");
}






