/* +---------------------------------------------------------------------------+
   |          The Mobile Robot Programming Toolkit (MRPT) C++ library          |
   |                                                                           |
   |                   http://mrpt.sourceforge.net/                            |
   |                                                                           |
   |   Copyright (C) 2005-2009  University of Malaga                           |
   |                                                                           |
   |    This software was written by the Perception and Robotics               |
   |      research group, University of Malaga (Spain).                        |
   |    Contact: Jose-Luis Blanco  <jlblanco@ctima.uma.es>                     |
   |                                                                           |
   |  This file is part of the MRPT project.                                   |
   |                                                                           |
   |     MRPT is free software: you can redistribute it and/or modify          |
   |     it under the terms of the GNU General Public License as published by  |
   |     the Free Software Foundation, either version 3 of the License, or     |
   |     (at your option) any later version.                                   |
   |                                                                           |
   |   MRPT is distributed in the hope that it will be useful,                 |
   |     but WITHOUT ANY WARRANTY; without even the implied warranty of        |
   |     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         |
   |     GNU General Public License for more details.                          |
   |                                                                           |
   |     You should have received a copy of the GNU General Public License     |
   |     along with MRPT.  If not, see <http://www.gnu.org/licenses/>.         |
   |                                                                           |
   +---------------------------------------------------------------------------+ */

#include <mrpt/core.h>
#include <mrpt/monoslam/CMonoSlam.h>
#include <mrpt/monoslam/CMonoSlamMotionModel.h>

using namespace std;
using namespace mrpt;
using namespace mrpt::utils;
using namespace mrpt::poses;
using namespace mrpt::slam;
using namespace mrpt::monoslam;
using namespace mrpt::math;
using namespace mrpt::gui;
using namespace mrpt::vision;
using namespace mrpt::hwdrivers;




/** This function return the position,zi (in pixels) of a feature yi
	  - Returns the pixel cooordinates (distorted) onto we expect to see this feature
	  - Performs a Projection from 3D (yi) to 2D (u,v) 
 */

void  CMonoSlam::hi(CVectorDouble &yinit, CMatrixDouble &zi, size_t indx)
{
    // Compute a single measurement
    // Points 3D in camera coordinates
    // R_wc is the rotation matrix

	CMatrixDouble R_wc;
	CQuaternionDouble q;
	mrpt::math::CVectorDouble r_wc_; //position of the camera the first time it observed the feature

	r_wc_.resize(3);

	m_xkk.extract_vector(3,q);			// q --> asociated quaternion to state vector
	m_xkk.extract_vector(0,r_wc_);		// r_wc --> world camera coordinates
	q.q_rotation_matrix(R_wc);			// From camera plane C to the plane in the world W
	CMatrixDouble R_cw(R_wc.inv());     // The inverse than above

	mrpt::math::CVectorDouble yi(3);		
 	yinit.extract_vector(0,yi);		        // yi = position (x,y,z) of the camera the first time it observed the feature
	const double theta(yinit[3]);			// theta = hx        = azimuth angle of the feature's vector
	const double phi(yinit[4]);				// phi =   hy        = elevation angle of the feature's vector
	const double rho(yinit[5]);				// rho = inverse distance of the feature (di = 1/rho)	

	mrpt::math::CVectorDouble mi(3);
    mi[0] =  cos(phi) * sin(theta);
	mi[1] = -sin(phi);
	mi[2] =  cos(phi) * cos(theta);

 	// hrl is the estimated vector between the camera center and the feature
	// hrl = R_cw * ( (yi - r_wc) * rho + mi );
	// pag 4 Real-Time Simultaneous Localisation and Mapping with a Single Camera
	// pag 2 Unified Inverse Deth Parametrization for Monocular SLAM

	mrpt::math::CVectorDouble hrl(3);
	CVectorDouble temp = yi - r_wc_;
	temp *= rho;
	temp += mi;

	// hrl = R_cw * temp
	R_cw.multiply_Ab(temp, hrl); 

    m_hrl_i.resize(3);
	m_hrl_i = hrl;

	// Is in front of the camera?
    if ((hrl[2] < 2.22e-16) && (rho > 0))
	{
        zi(0,0) = -1;	// It's out of sight of the camera
		zi(0,1)	= -1;
	}

    // Image coordinates
	TPixelCoordf   uv_d; //distort coordinates

	//MAAA: OLD methods
	//TPixelCoordf   uv_u;    //undistort coordinates
    //hu( TPoint3D(hrl[0],hrl[1],hrl[2]) , uv_u);  // projection of the point zero_yi (hrl)
	//Add distortion
	//cam.distort_a_point( uv_u , uv_d );    //uv_d --> pixel coordinates (distorted)

	// Project point with distortion, in one step:
	cam.project_3D_point(TPoint3D(hrl[0],hrl[1],hrl[2]), uv_d);

	zi.set_unsafe(0,0,  uv_d.x );
	zi.set_unsafe(0,1,  uv_d.y );
		
   //std::cout<<"[Mono-Slam] hi  zi   : " << zi   << "\n";
   //std::cout<<"[Mono-Slam] Yi       : " << yinit << "\n";
   //std::cout<<"[Mono-Slam] Camera   x: "<<r_wc_[0]  <<"		y: " << r_wc_[1] <<"		z: "<<r_wc_[2] << "\n";
   //std::cout<<"[Mono-Slam] 3D  hRL  x: "<<hrl[0]  <<"	y: " << hrl[1] <<"	z: "<<hrl[2] << "\n";
 
}


/******************************************************************************************************************************/
// OLD and NOT USED
// Project a 3D point into undistorted pixel coordinates
void  CMonoSlam::hu(const mrpt::math::TPoint3D &XYZ, mrpt::vision::TPixelCoordf &out_pixel )
{

// OLD and NOT USED

   // This turns the position vector into an undistorted image location (with the
   // origin at the centre)
   // Use -Fkuv to swap the image co-ordinates (0,0) at the top to
   // camera co-ordinates (0,0) at the bottom

	ASSERT_(XYZ.z!=0)
	out_pixel.x = cam.cam_cx() + (XYZ.x / XYZ.z)*cam.cam_fx();  // JL: I think the correct sign is "+"
	out_pixel.y = cam.cam_cy() + (XYZ.y / XYZ.z)*cam.cam_fy();  //  for img. origin at the top-left

   //std::cout<<"[Mono-Slam] hu  out_pixel  x: "<<out_pixel.x <<"		y: "<<out_pixel.y << "\n";

}




//--------------------------------------------





/******************************************************************************************************************************/
CMatrixDouble CMonoSlam::calculate_ray(const float &col,const float &row)
{
//  col = column (in distorted pixels) of the feature
//	row = row (in distorted pixels) of the feature

	const double fku = cam.cam_fx();
	const double fkv = cam.cam_fy();
	const double U0  = cam.cam_cx();
	const double V0  = cam.cam_cy();

  //Elimina la distorsion de la posicion de la caracteristica
	TPixelCoordf p_undist;
	cam.undistort_point( TPixelCoordf(col,row), p_undist );

	CQuaternionDouble q_WR;		//quaternion de rotacion
	m_xkk.extract_vector(3,q_WR);

	// Pinhole camera model. Coordinates u,v (x,y,z in homogeneous) predicted for the feature onto the image
	double h_LR_x=(U0-p_undist.x)/fku;
	double h_LR_y=(V0-p_undist.y)/fkv;
	double h_LR_z=1.0;

	// Ray from feature to optic center of the camera
	CMatrixDouble h_LR(3,1);
	h_LR(0,0)=h_LR_x;
	h_LR(1,0)=h_LR_y;
	h_LR(2,0)=h_LR_z;

	// Convert the quaternion to a rotation matrix and multiply by the point h_LR
	//  n is for computing angles theta & phi of the feature wrt global world axis
	CMatrixDouble rotation_matrix;
	q_WR.q_rotation_matrix(rotation_matrix);
	CMatrixDouble n;
	n.setSize(3,1);
	n=rotation_matrix*h_LR;     // coords u,v,1 of the camera -> u',v',1' in the world

	n = n / n.norm();

	return n;
}
/******************************************************************************************************************************/

mrpt::math::CVectorDouble  CMonoSlam::hinv(CMatrixDouble h_LR_rot)
{
//  (col,row) = (distorted) coordinates of the feature point


	mrpt::math::CVectorDouble newFeature;
	newFeature.resize(feature_size);

	CVectorDouble r_W;	//posicion de la camara
	r_W.resize(3);

	m_xkk.extract_vector(0,r_W);
	newFeature.insert_vector(0,r_W);

	newFeature[0] += h_LR_rot(0,0); //Feature Wolrd Pos
	newFeature[1] += h_LR_rot(1,0);
	newFeature[2] += h_LR_rot(2,0);

	float nx =  h_LR_rot(0,0);     //Feature Local Pos relative to camera
	float ny =  h_LR_rot(1,0);
	float nz =  h_LR_rot(2,0);

	// (Antonio Jose) Warning! In the paper on inverse parameterization theta and phi are in the wrong order.

	newFeature[3]=atan2(nx,nz);
	newFeature[4]=atan2(static_cast<double>(-ny),static_cast<double>(hypot(nx,nz)));
	newFeature[5]= options.lambda_init;

	return newFeature;
}
/******************************************************************************************************************************/
void  CMonoSlam::addAFeatureCov_newPar(CMatrixDouble &XYZ_w,float &col, float &row)
{
//	pkk --> p(k|k)
//	XYZ_w --> ray from camera origin (focus point) to the feature, in world coordinates

	CQuaternionDouble q_wc;		// rotation quaternion
	m_xkk.extract_vector(3,q_wc);
	CMatrixDouble R_wc(3,3), R_cw(3,3), XYZ_c(3,1);
	q_wc.q_rotation_matrix(R_wc);
	R_cw = R_wc.inv();
	XYZ_c = R_cw * XYZ_w;

	double X_w = XYZ_w(0,0);
	double Y_w = XYZ_w(1,0);
	double Z_w = XYZ_w(2,0);

	// Jacobians - from page 4 in "A visual compass based on SLAM"

	// gw is the ray from the origin to the feature.
	CMatrixDouble dgw_dqwr(3,4);
	dgw_dqwr = dRq_times_a_by_dq( q_wc, XYZ_c );

	CMatrixDouble dtheta_dgw;
	dtheta_dgw.setSize(1,3);
	dtheta_dgw(0,0) = Z_w / (X_w * X_w + Z_w * Z_w);
	dtheta_dgw(0,2) = -X_w / (X_w * X_w + Z_w * Z_w);

	CMatrixDouble dphi_dgw;
	dphi_dgw.setSize(1,3);
	dphi_dgw(0,0) = ( X_w * Y_w )/( (X_w*X_w + Y_w*Y_w + Z_w*Z_w )*sqrt(X_w*X_w + Z_w*Z_w) );
	dphi_dgw(0,1) = -sqrt( X_w*X_w + Z_w*Z_w )/( X_w*X_w + Y_w*Y_w + Z_w*Z_w );
	dphi_dgw(0,2) = ( Z_w*Y_w )/( (X_w*X_w + Y_w*Y_w + Z_w*Z_w )*sqrt(X_w*X_w + Z_w*Z_w) );

	CMatrixDouble dtheta_dqwr;
	dtheta_dqwr.setSize(1,4);
	dtheta_dqwr = dtheta_dgw * dgw_dqwr;

	CMatrixDouble dphi_dqwr;
	dphi_dqwr.setSize(1,4);
	dphi_dqwr = dphi_dgw*dgw_dqwr;

	CMatrixDouble dy_dqwr;
	dy_dqwr.zeros(feature_size,4);
	dy_dqwr.insertMatrix(3,0,dtheta_dqwr);
	dy_dqwr.insertMatrix(4,0,dphi_dqwr);

	CMatrixDouble dy_drw;
	dy_drw.zeros(feature_size,3);
	dy_drw(0,0) = 1.0;
	dy_drw(1,1) = 1.0;
	dy_drw(2,2) = 1.0;

	CMatrixDouble dy_dxv;
	dy_dxv.zeros(feature_size,state_size);
	dy_dxv.insertMatrix(0,0,dy_drw);
	dy_dxv.insertMatrix(0,3,dy_dqwr);

	//calculate dy_dhd
	CMatrixDouble dy_dhd;
	CMatrixDouble dyprima_dgw;
	dyprima_dgw.zeros(5,3);
	dyprima_dgw.insertMatrix(3,0,dtheta_dgw);
	dyprima_dgw.insertMatrix(4,0,dphi_dgw);

	//CMatrixDouble dgc_dhu(3,2);
	//dgc_dhu(0,0) = 1.0/ cam.cam_fx();
	//dgc_dhu(1,1) = 1.0/ cam.cam_fy();
	//CMatrixDouble dhu_dhd(2,3);
	//cam.jacob_undistor( TPixelCoordf(col,row), dhu_dhd );
	
	//added by MAAA:
	CMatrixDouble dgc_dhu_dhu_dhd;
    cam.jacobian_unproject_with_distortion(TPixelCoordf(col,row), dgc_dhu_dhu_dhd );
	
	CMatrixDouble dyprima_dhd(5,2);
	dyprima_dhd = dyprima_dgw * R_wc * dgc_dhu_dhu_dhd/* dgc_dhu * dhu_dhd*/;
	dy_dhd.zeros(feature_size,3);
	dy_dhd.insertMatrix(0,0,dyprima_dhd);
	dy_dhd(5,2)=1.0;

	//std::cout << "[Mono-Slam]  dgc_dhu_dhu_dhd:"<< dgc_dhu_dhu_dhd << "\n"; 
	//std::cout << "[Mono-Slam]  dyprima_dhd:"<< dyprima_dhd << "\n"; 
  
	//Jacobian
	CMatrixDouble I,J;
	I.unit(m_pkk.getColCount());
	J.zeros(m_pkk.getRowCount()+ feature_size,m_pkk.getColCount()+3);
	J.insertMatrix(0,0,I);
	J.insertMatrix(m_pkk.getRowCount(),0,dy_dxv);
	J.insertMatrix(m_pkk.getRowCount(),m_pkk.getColCount(),dy_dhd);

	//if (m_savedatas)
			//dy_dhd.saveToTextFile("./monoSlamOut/dy_dhd.txt");

	CMatrixDouble Ri;
	Ri.unit(2);
	Ri *= square( options.sigma_image_noise );

	// Add element (2,2):
	Ri.setSize(3,3);
	Ri(2,2) = square( options.std_lambda );

	//CMatrixDouble Padd;
	//Padd.zeros(3,3);
	//Padd.insertMatrix(0,0,Ri);
	//Padd(2,2) = square( mmm.get_std_lambda() );

	// Expand Pkk cov matrix for one more  feature:
	CMatrixDouble temp(m_pkk.getRowCount(),m_pkk.getColCount());
	//temp = m_pkk;
	//m_pkk.zeros(m_pkk.getRowCount()+3,m_pkk.getColCount()+3);
	//m_pkk.insertMatrix(0,0,temp);

	m_pkk.setSize( size(m_pkk,1)+3,  size(m_pkk,2)+3 );  // This pads with zeroes.
	m_pkk.insertMatrix( size(m_pkk,1)-3,size(m_pkk,1)-3, Ri);

	const CMatrixDouble m_pkk_copy(m_pkk);

	J.multiply_HCHt(m_pkk_copy,m_pkk);  // m_pkk = J * m_pkk * ~J;

    //CMatrixDouble pKKnew;
	//m_pkk.extractMatrix( size(m_pkk,1)-3,size(m_pkk,2)-3,3,3, pKKnew);
	//std::cout << "[Mono-Slam]  pKKnew: \n"<< pKKnew  << "\n"; 
	//std::cout << "[Mono-Slam]  Ri: \n"<< Ri  << "\n"; 
	//cin.get();
	//std::cout << "[Mono-Slam]  J:"<< J << "\n"; 
	//std::cout << "[Mono-Slam]  m_pkk_copy:"<< m_pkk_copy << "\n"; 
	//std::cout << "[Mono-Slam]  m_pkk:"<< m_pkk  << "\n"; 

}




